from weaveio.basequery.common import NotYetImplementedError, FrozenQuery


class BranchLogic(FrozenQuery):
    def __init__(self, parent):
        super(BranchLogic, self).__init__(parent=parent)
        self.branches = []

    def simplify(self):
        pass




from weaveio.data import Data
from weaveio.opr3.l1files import RawFile, L1SingleFile, L1StackFile, L1SuperStackFile, L1SuperTargetFile
from weaveio.opr3.l2files import StackL2File, SuperStackL2File


class OurData(Data):
    filetypes = [RawFile, L1SingleFile, L1StackFile, L1SuperStackFile, L1SuperTargetFile, StackL2File, SuperStackL2File]


import urllib

import json
from datetime import datetime

import logging

from .__version__ import __version__

import atexit
from typing import Callable

from boxing import boxing
from colored import fore, style
from pkg_info import get_pkg_info
from semver import compare


def parse_changes(description):
    log = description.split('machine-readable-change-log\n###########################\n')[-1]
    return log


def get_other_descriptions(name, version):
    return json.loads(urllib.request.urlopen(f'https://pypi.org/pypi/{name}/{version}/json', timeout=3).read())['info']['description']


class UpdateNotify(object):
    def __init__(self, name: str, version: str, **opts):
        self.name: str = name
        self.version: str = version
        self.pkg = get_pkg_info(self.name)
        self.is_updated: bool = self.is_latest_version()
        self.callback: Callable = opts.get('callback')
        self.message: str = opts.get('message')
        self.defer: bool = bool(opts.get('defer', False))

    def is_latest_version(self) -> bool:
        self.latest = self.pkg.version
        return True if compare(self.version, self.latest) >= 0 else False

    def render_changes(self):
        releases = [(k, datetime.strptime(v[0]['upload_time'], '%Y-%m-%dT%H:%M:%S')) for k, v in self.pkg.raw_data['releases'].items()]
        releases.sort(key=lambda x: x[1])
        release_names, _ = zip(*releases)
        changes = [parse_changes(get_other_descriptions(self.pkg.name, v)) for v in release_names]
        newer_changes = changes[release_names.index(self.pkg.version):]
        return '\n'.join(newer_changes), len(newer_changes)

    def notify(self) -> None:
        if self.is_updated:
            return
        if self.callback and callable(self.callback):
            action, arg = self.callback, None
        elif self.message:
            action, arg = print, self.message
        else:
            action, arg = print, self.default_message()
        if self.defer:
            atexit.register(action, arg) if arg else atexit.register(action)
        else:
            action(arg) if arg else action()

    def default_message(self) -> str:
        changes, nchanges = self.render_changes()
        version = fore.GREY_53 + self.version + style.RESET
        latest = fore.LIGHT_GREEN + self.latest + style.RESET
        command = fore.LIGHT_BLUE + 'pip install -U ' + self.name + style.RESET
        nchanges = fore.LIGHT_GREEN + str(nchanges) + style.RESET
        return boxing(f'Update available {version} → {latest} ({nchanges} new changes)\n' +
                      f'Run {command} to update\n'
                      f'{changes}')


try:
    UpdateNotify('weaveio', __version__).notify()
except Exception as e:
    logging.exception('There was a problem in alerting you to updated versions of the weaveio library...', exc_info=True)



from pathlib import Path
from typing import Union, List, Dict

from astropy.io import fits
from astropy.table import Table

from weaveio.file import File, PrimaryHDU, TableHDU
from weaveio.graph import Graph
from weaveio.hierarchy import Multiple, unwind, collect
from weaveio.opr3.hierarchy import ClassificationTable, GalaxyTable, GalaxySpectrum, ClassificationSpectrum, APS, L1SpectrumRow, FibreTarget, L2_FTYPES, L2_DTYPES, OB, OBSpec
from weaveio.opr3.l1files import L1File, L1SuperStackFile, L1StackFile
from weaveio.writequery import CypherData


# make new types based on stack level and type of data
L2_TYPES = {}
for ftype in L2_FTYPES:
    for dtype in L2_DTYPES:
        name = ftype.__name__.replace('L2', '') + dtype.__name__
        L2_TYPES[name] = type(name, (ftype, dtype), {'is_template': False})


class MissingDataError(Exception):
    pass


def filter_products_from_table(table: Table, maxlength: int) -> Table:
    columns = []
    for i in table.colnames:
        value = table[i]
        if len(value.shape) == 2:
            if value.shape[1] > maxlength:
                continue
        columns.append(i)
    return table[columns]


class L2File(File):
    is_template = True
    match_pattern = '*aps.fits'
    produces = [ClassificationTable, GalaxyTable]#, ClassificationSpectrum, GalaxySpectrum]
    corresponding_hdus = ['class_table', 'galaxy_table']#, 'class_spectra', 'galaxy_spectra']
    parents = [Multiple(L1File, 2, 3)]
    hdus = {'primary': PrimaryHDU, 'fibtable': TableHDU,
            'class_spectra': TableHDU,
            'stellar_spectra_ferre': TableHDU, 'stellar_spectra_rvs': TableHDU,
            'galaxy_spectra': TableHDU,
            'class_table': TableHDU,
            'stellar_table': TableHDU,
            'stellar_table_rvs': TableHDU,
            'galaxy_table': TableHDU}
    recommended_batchsize = 600

    @classmethod
    def length(cls, path):
        hdus = fits.open(path)
        names = [i.name for i in hdus]
        return len(hdus[names.index('CLASS_SPECTRA')].data)

    @classmethod
    def query_structure(cls, path, graph):
        header = cls.read_header(path)
        runids = list(map(int, header['RUN'].split('+')))
        result = graph.execute('UNWIND $runids as runid MATCH (run:Run {runid:runid}) '
                               'MATCH (run)<-[*]-(ob:OB)'
                               'MATCH (ob)<--(obspec:OBSpec) '
                               'return collect(distinct ob.obid) as obids, collect(distinct obspec.xml) as xmls',
                               runids=runids).to_table()

        if not len(result):
            raise MissingDataError(f"Data for runs {runids} has not been found in the database. "
                                   f"Unable to decide what type of file {path} is.")
        obids, xmls = result[0][0], result[0][1]
        return obids, xmls

    @classmethod
    def parse_fname(cls, header, fname, instantiate=True) -> List[L1File]:
        ftype_dict = {'stacked': L1StackFile, 'stack': L1StackFile,
                      'superstack': L1SuperStackFile, 'superstacked': L1SuperStackFile}
        split = fname.name.replace('.aps.fits', '').replace('.aps.fit', '').split('_')
        runids = []
        ftypes = []
        for i in split:
            try:
                runids.append(int(i))
            except ValueError:
                ftypes.append(str(i))
        if len(ftypes) == 1:
            ftypes = [ftypes[0]] * len(runids)  # they all have the same type if there is only one mentioned
        assert len(ftypes) == len(runids), "error parsing runids/types from fname"
        assert all(int(i) in runids for i in header['RUN'].split('+')), "fname runids and header runids do not match"
        files = []
        for ftype, runid in zip(ftypes, runids):
            ftype_cls = ftype_dict[ftype]
            fname = ftype_cls.fname_from_runid(runid)
            if instantiate:
                files.append(ftype_cls.find(fname=fname))
            else:
                files.append((ftype_cls, fname))
        return files

    @classmethod
    def find_shared_hierarchy(cls, path: Path) -> Dict:
        raise NotImplementedError

    @classmethod
    def read_header(cls, path):
        return fits.open(path)[0].header

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None):
        fname = Path(fname)
        directory = Path(directory)
        path = directory / fname
        header = cls.read_header(path)
        l1files = cls.parse_fname(header, fname)
        hdu_nodes, file = cls.read_hdus(directory, fname, l1files=l1files)
        astropyhdulist = fits.open(path)
        aps = APS(apsvers=header['APSVERS'])
        hierarchies = cls.find_shared_hierarchy(path)
        for name in cls.corresponding_hdus:
            cls.make_data_rows(name, slc, l1files, file, astropyhdulist, hdu_nodes, aps, **hierarchies)
        return file

    @classmethod
    def read_one_hdu_l2data(cls, hdus, hduname, slc=None):
        slc = slice(None) if slc is None else slc
        names = [i.name.lower().strip() for i in hdus]
        table = Table(hdus[names.index(hduname)].data)[slc]
        if len(table.colnames):
            table.rename_columns(table.colnames, [i.lower() for i in table.colnames])
            table['spec_index'] = range(len(table))
        table = filter_products_from_table(table, 10)  # removes huge arrays that are best kept in binary files
        data = CypherData(table, hduname)
        return data

    @classmethod
    def make_data_rows(cls, hduname, slc, l1files, file, astropyhdulist, hdus, aps, **hierarchies):
        row_type = cls.produces[cls.corresponding_hdus.index(hduname)]
        table = cls.read_one_hdu_l2data(astropyhdulist, hduname, slc)
        l1filenames = CypherData([l.fname for l in l1files], 'l1fnames')
        with unwind(table) as row:
            with unwind(l1filenames) as l1fname:
                spectrum = L1SpectrumRow.find(sourcefile=l1fname, nrow=row['nspec'])
            spectra = collect(spectrum)
            spectra_list = [spectra[i] for i, _ in enumerate(l1files)]
            fibretarget = FibreTarget.find(anonymous_children=[spectra[0]])
            l2row = row_type(sourcefile=file.fname, nrow=row['nspec'],
                             tables=row, l1spectrumrows=spectra_list,
                             fibretarget=fibretarget, aps=aps, **hierarchies)
            l2row.attach_products(file, index=row['spec_index'], **hdus)
        l2rows = collect(l2row)
        return l2rows


class StackL2File(L2File):
    produces = [L2_TYPES['StackClassificationTable'], L2_TYPES['StackGalaxyTable']]#,
                #L2_TYPES['StackClassificationSpectrum'], L2_TYPES['StackGalaxySpectrum']]
    corresponding_hdus = ['class_table', 'galaxy_table',]# 'class_spectra', 'galaxy_spectra']

    @classmethod
    def match_file(cls, directory: Union[Path, str], fname: Union[Path, str], graph: Graph):
        fname = Path(fname)
        directory = Path(directory)
        path = directory / fname
        if not super().match_file(directory, fname, graph):
            return False
        obids, xmls = cls.query_structure(path, graph)
        if len(obids) == 1:
            return True
        return False

    @classmethod
    def find_shared_hierarchy(cls, path) -> Dict:
        header = cls.read_header(path)
        return {'ob': OB.find(obid=header['OBID'])}


class SuperStackL2File(L2File):
    produces = [L2_TYPES['SuperStackClassificationTable'], L2_TYPES['SuperStackGalaxyTable']]#,
                #L2_TYPES['SuperStackClassificationSpectrum'], L2_TYPES['SuperStackGalaxySpectrum']]
    corresponding_hdus = ['class_table', 'galaxy_table',]# 'class_spectra', 'galaxy_spectra']

    @classmethod
    def match_file(cls, directory: Union[Path, str], fname: Union[Path, str], graph: Graph):
        fname = Path(fname)
        directory = Path(directory)
        path = directory / fname
        if not super().match_file(directory, fname, graph):
            return False
        obids, xmls = cls.query_structure(path, graph)
        if len(obids) == 1:
            return False
        elif len(xmls) == 1:
            return True
        else:
            return False

    @classmethod
    def find_shared_hierarchy(cls, path) -> Dict:
        header = cls.read_header(path)
        return {'obspec': OBSpec.find(xml=str(header['cat-name']))}



import time
from functools import reduce
from itertools import product
from operator import or_

import networkx
import numpy as np
import logging
from pathlib import Path
from typing import Union, List, Tuple, Type, Dict
import pandas as pd
import re

import networkx as nx
import py2neo
from tqdm import tqdm

from weaveio.address import Address
from weaveio.basequery.handler import Handler, defaultdict
from weaveio.basequery.tree import BranchHandler, TraversalPath
from weaveio.graph import Graph
from weaveio.writequery import Unwind
from weaveio.hierarchy import Multiple, Hierarchy, Indexed
from weaveio.file import File, HDU

CONSTRAINT_FAILURE = re.compile(r"already exists with label `(?P<label>[^`]+)` and property "
                                r"`(?P<idname>[^`]+)` = (?P<idvalue>[^`]+)$", flags=re.IGNORECASE)

def process_neo4j_error(data: 'Data', file: File, msg):
    matches = CONSTRAINT_FAILURE.findall(msg)
    if not len(matches):
        return  # cannot help
    label, idname, idvalue = matches[0]
    # get the node properties that already exist
    extant = data.graph.neograph.evaluate(f'MATCH (n:{label} {{{idname}: {idvalue}}}) RETURN properties(n)')
    fname = data.graph.neograph.evaluate(f'MATCH (n:{label} {{{idname}: {idvalue}}})-[*]->(f:File) return f.fname limit 1')
    idvalue = idvalue.strip("'").strip('"')
    file.data = data
    obj = [i for i in data.hierarchies if i.__name__ == label][0]
    instance_list = getattr(file, obj.plural_name)
    new = {}
    if not isinstance(instance_list, (list, tuple)):  # has an unwind table object
        new_idvalue = instance_list.identifier
        if isinstance(new_idvalue, Unwind):
            # find the index in the table and get the properties
            filt = (new_idvalue.data == idvalue).iloc[:, 0]
            for k in extant.keys():
                if k == 'id':
                    k = idname
                value = getattr(instance_list, k, None)
                if isinstance(value, Unwind):
                    table = value.data.where(pd.notnull(value.data), 'NaN')
                    new[k] = str(table[k][filt].values[0])
                else:
                    new[k] = str(value)
        else:
            # if the identifier of this object is not looping through a table, we cant proceed
            return
    else:  # is a list of non-table things
        found = [i for i in instance_list if i.identifier == idvalue][0]
        for k in extant.keys():
            value = getattr(found, k, None)
            new[k] = value
    comparison = pd.concat([pd.Series(extant, name='extant'), pd.Series(new, name='to_add')], axis=1)
    filt = comparison.extant != comparison.to_add
    filt &= ~comparison.isnull().all(axis=1)
    where_different = comparison[filt]
    logging.exception(f"The node (:{label} {{{idname}: {idvalue}}}) tried to be created twice with different properties.")
    logging.exception(f"{where_different}")
    logging.exception(f"filenames: {fname}, {file.fname}")


def get_all_subclasses(cls):
    all_subclasses = []
    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))
    return all_subclasses


def find_children_of(parent):
    hierarchies = get_all_subclasses(Hierarchy)
    children = set()
    for h in hierarchies:
        if len(h.parents):
            if any(p is parent if isinstance(p, type) else p.node is parent for p in h.parents):
                children.add(h)
    return children



class Data:
    filetypes = []

    def read_in_filetype(self, filetype: File):
        todo = set(filetype.produces + list(filetype.hdus.values()))
        todo.add(filetype)
        while len(todo):
            thing = todo.pop()
            if not thing.is_template:
                self.hierarchies.add(thing)
                for hier in thing.parents:
                    if not thing.is_template:
                        if isinstance(hier, Multiple):
                            todo.add(hier.node)
                        else:
                            todo.add(hier)

    def __init__(self, rootdir: Union[Path, str], host: str = 'host.docker.internal', port=11002, write=False,
                 password=None, user=None):
        self.branch_handler = BranchHandler()
        self.handler = Handler(self)
        self.host = host
        self.port = port
        self.write_allowed = write
        self._graph = None
        self.password = password
        self.user = user
        self.filelists = {}
        self.rootdir = Path(rootdir)
        self.address = Address()
        self.hierarchies = set()
        for f in self.filetypes:
            self.read_in_filetype(f)
        self.class_hierarchies = {h.__name__: h for h in self.hierarchies}
        self.singular_hierarchies = {h.singular_name: h for h in self.hierarchies}  # type: Dict[str, Type[Hierarchy]]
        self.plural_hierarchies = {h.plural_name: h for h in self.hierarchies if h.plural_name != 'graphables'}
        self.factor_hierarchies = defaultdict(list)
        for h in self.hierarchies:
            for f in getattr(h, 'factors', []):
                self.factor_hierarchies[f.lower()].append(h)
            if h.idname is not None:
                self.factor_hierarchies[h.idname].append(h)
        self.factor_hierarchies = dict(self.factor_hierarchies)  # make sure we always get keyerrors when necessary!
        self.factors = set(self.factor_hierarchies.keys())
        self.plural_factors =  {f.lower() + 's': f.lower() for f in self.factors}
        self.singular_factors = {f.lower() : f.lower() for f in self.factors}
        self.singular_idnames = {h.idname: h for h in self.hierarchies if h.idname is not None}
        self.plural_idnames = {k+'s': v for k,v in self.singular_idnames.items()}
        self.make_relation_graph()
        self.make_dually_directed_graph()

    def make_dually_directed_graph(self):
        """
        Turns a directed graph with one with edges in both directions
        With an edge of (a)-[number=2, multiple=True]>(b), b has 2 of a, a has <unknown> of b
        """
        for a, b in list(self.relation_graph.edges):
            attrs = self.relation_graph.edges[(a, b)]
            if attrs.get('belongs_to', False):
                multiplicity = False
                number = 1
            else:
                multiplicity = True
                number = None
            self.relation_graph.add_edge(b, a, multiplicity=multiplicity, number=number, inverted=True)


    def write(self, collision_manager='track&flag'):
        if self.write_allowed:
            return self.graph.write(collision_manager)
        raise IOError(f"You have not allowed write operations in this instance of data (write=False)")

    def is_unique_factor(self, name):
        return len(self.factor_hierarchies[name]) == 1

    @property
    def graph(self):
        if self._graph is None:
            d = {}
            if self.password is not None:
                d['password'] = self.password
            if self.user is not None:
                d['user'] = self.user
            self._graph = Graph(host=self.host, port=self.port, write=self.write, **d)
        return self._graph

    def make_relation_graph(self):
        self.relation_graph = nx.DiGraph()
        d = list(self.singular_hierarchies.values())
        while len(d):
            h = d.pop()
            try:
                is_file = issubclass(h, File)
            except:
                is_file = False
            self.relation_graph.add_node(h.singular_name, is_file=is_file, is_hdu=issubclass(h, HDU),
                                         factors=h.factors+[h.idname], idname=h.idname)
            for parent in h.parents:
                multiplicity = isinstance(parent, Multiple)
                if multiplicity:
                    if parent.maxnumber == parent.minnumber:
                        number = parent.maxnumber
                        numberlabel = f'={number}'
                    else:
                        number = None
                        if (parent.minnumber is None or parent.minnumber == 0) and parent.maxnumber is None:
                            numberlabel = 'any'
                        elif (parent.minnumber is None or parent.minnumber == 0) and parent.maxnumber is not None:
                            numberlabel = f'<= {parent.maxnumber}'
                        elif (parent.minnumber is not None and parent.minnumber > 0) and parent.maxnumber is None:
                            numberlabel = f'>={parent.minnumber}'
                        else:
                            numberlabel = f'{parent.minnumber} - {parent.maxnumber}'
                    parent = parent.node
                else:
                    number = 1
                    numberlabel = f'={number}'
                subclasses = [parent] + get_all_subclasses(parent)
                for subclass in subclasses:
                    if not subclass.is_template:
                        self.relation_graph.add_node(subclass.singular_name, is_file=is_file, is_hdu=issubclass(subclass, HDU),
                                                     factors=subclass.factors+[subclass.idname],
                                                     idname=subclass.idname)
                        self.relation_graph.add_edge(subclass.singular_name, h.singular_name, belongs_to=subclass in h.belongs_to,
                                                     multiplicity=multiplicity, number=number,
                                                     label=numberlabel)
                        d.append(subclass)
            for hier in h.produces:
                self.relation_graph.add_edge(h.singular_name, hier.singular_name, belongs_to=h in hier.belongs_to,
                                             multiplicity=False, number=1, index=False, name='file',
                                             label='=1')
                for product_name in hier.products:
                    if isinstance(product_name, Indexed):
                        index = True
                        product_name = product_name.name
                    else:
                         index = False
                    product = h.hdus[product_name]
                    self.relation_graph.add_edge(product.singular_name, hier.singular_name, belongs_to=product in hier.belongs_to,
                                                 multiplicity=False, number=1, index=index, name=product_name,
                                                 label='=1')

    def make_constraints_cypher(self):
        return [hierarchy.make_schema() for hierarchy in self.hierarchies]

    def apply_constraints(self):
        if not self.write_allowed:
            raise IOError(f"Writing is not allowed")
        for q in tqdm(self.make_constraints_cypher(), desc='applying constraints'):
            self.graph.neograph.run(q)

    def drop_all_constraints(self):
        if not self.write_allowed:
            raise IOError(f"Writing is not allowed")
        self.graph.neograph.run('CALL apoc.schema.assert({},{},true) YIELD label, key RETURN *')

    def get_extant_files(self):
        return self.graph.execute("MATCH (f:File) RETURN DISTINCT f.fname").to_series(dtype=str).values.tolist()

    def raise_collisions(self):
        """
        returns the properties that would have been overwritten in nodes and relationships.
        """
        node_collisions = self.graph.execute("MATCH (c: _Collision) return c { .*}").to_data_frame()
        rel_collisions = self.graph.execute("MATCH ()-[c: _Collision]-() return c { .*}").to_data_frame()
        return node_collisions, rel_collisions

    def read_files(self, *paths: Union[Path, str], collision_manager='ignore', batch_size=None, halt_on_error=False) -> pd.DataFrame:
        """
        Read in the files given in `paths` to the database.
        `collision_manager` is the method with which the database deals with overwriting data.
        Values of `collision_manager` can be {'ignore', 'overwrite', 'track&flag'}.
        track&flag will have the same behaviour as ignore but places the overlapping data in its own node for later retrieval.
        :return
            statistics dataframe
        """
        batches = []
        for path in paths:
            path = Path(path)
            matches = [f for f in self.filetypes if f.match_file(self.rootdir, path.relative_to(self.rootdir), self.graph)]
            if len(matches) > 1:
                raise ValueError(f"{path} matches more than 1 file type: {matches} with `{[m.match_pattern for m in matches]}`")
            filetype = matches[0]
            filetype_batch_size = filetype.recommended_batchsize if batch_size is None else batch_size
            slices = filetype.get_batches(path, filetype_batch_size)
            batches += [(filetype, path.relative_to(self.rootdir), slc) for slc in slices]
        elapsed_times = []
        stats = []
        timestamps = []
        bar = tqdm(batches)
        for filetype, fname, slc in bar:
            bar.set_description(f'{fname}[{slc.start}:{slc.stop}]')
            with self.write(collision_manager) as query:
                filetype.read(self.rootdir, fname, slc)
            cypher, params = query.render_query()
            start = time.time()
            try:
                results = self.graph.execute(cypher, **params)
            except py2neo.database.work.ClientError as e:
                logging.exception('ClientError:', exc_info=True)
                if halt_on_error:
                    raise e
                print(e)
            stats.append(results.stats())
            timestamps.append(results.evaluate())
            elapsed_times.append(time.time() - start)
        df = pd.DataFrame(stats)
        df['timestamp'] = timestamps
        df['elapsed_time'] = elapsed_times
        if len(batches):
            _, df['fname'], slcs = zip(*batches)
            df['batch_start'], df['batch_end'] = zip(*[(i.start, i.stop) for i in slcs])
        else:
            df['fname'], df['batch_start'], df['batch_end'] = [], [], []
        return df.set_index(['fname', 'batch_start', 'batch_end'])

    def read_directory(self, *filetype_names, collision_manager='ignore', skip_extant_files=True, halt_on_error=False):
        filelist = []
        extant_fnames = self.get_extant_files() if skip_extant_files else []
        print(f'Skipping {len(extant_fnames)} extant files (use skip_extant_files=False to go over them again)')
        if len(filetype_names) == 0:
            filetypes = self.filetypes
        else:
            filetypes = [f for f in self.filetypes if f.singular_name in filetype_names]
        for filetype in filetypes:
            filelist += [i for i in self.rootdir.rglob(filetype.match_pattern) if str(i.relative_to(self.rootdir)) not in extant_fnames]
        return self.read_files(*filelist, collision_manager=collision_manager, halt_on_error=halt_on_error)

    def _validate_one_required(self, hierarchy_name):
        hierarchy = self.singular_hierarchies[hierarchy_name]
        parents = [h for h in hierarchy.parents]
        qs = []
        for parent in parents:
            if isinstance(parent, Multiple):
                mn, mx = parent.minnumber, parent.maxnumber
                b = parent.node.__name__
            else:
                mn, mx = 1, 1
                b = parent.__name__
            mn = 0 if mn is None else mn
            mx = 9999999 if mx is None else mx
            a = hierarchy.__name__
            q = f"""
            MATCH (n:{a})
            WITH n, SIZE([(n)<-[]-(m:{b}) | m ])  AS nodeCount
            WHERE NOT (nodeCount >= {mn} AND nodeCount <= {mx})
            RETURN "{a}", "{b}", {mn} as mn, {mx} as mx, n.id, nodeCount
            """
            qs.append(q)
        if not len(parents):
            qs = [f"""
            MATCH (n:{hierarchy.__name__})
            WITH n, SIZE([(n)<-[:IS_REQUIRED_BY]-(m) | m ])  AS nodeCount
            WHERE nodeCount > 0
            RETURN "{hierarchy.__name__}", "none", 0 as mn, 0 as mx, n.id, nodeCount
            """]
        dfs = []
        for q in qs:
            dfs.append(self.graph.neograph.run(q).to_data_frame())
        df = pd.concat(dfs)
        return df

    def _validate_no_duplicate_relation_ordering(self):
        q = """
        MATCH (a)-[r1]->(b)<-[r2]-(a)
        WHERE TYPE(r1) = TYPE(r2) AND r1.order <> r2.order
        WITH a, b, apoc.coll.union(COLLECT(r1), COLLECT(r2))[1..] AS rs
        RETURN DISTINCT labels(a), a.id, labels(b), b.id, count(rs)+1
        """
        return self.graph.neograph.run(q).to_data_frame()

    def _validate_no_duplicate_relationships(self):
        q = """
        MATCH (a)-[r1]->(b)<-[r2]-(a)
        WHERE TYPE(r1) = TYPE(r2) AND PROPERTIES(r1) = PROPERTIES(r2)
        WITH a, b, apoc.coll.union(COLLECT(r1), COLLECT(r2))[1..] AS rs
        RETURN DISTINCT labels(a), a.id, labels(b), b.id, count(rs)+1
        """
        return self.graph.neograph.run(q).to_data_frame()

    def validate(self):
        duplicates = self._validate_no_duplicate_relationships()
        print(f'There are {len(duplicates)} duplicate relations')
        if len(duplicates):
            print(duplicates)
        duplicates = self._validate_no_duplicate_relation_ordering()
        print(f'There are {len(duplicates)} relations with different orderings')
        if len(duplicates):
            print(duplicates)
        schema_violations = []
        for h in tqdm(list(self.singular_hierarchies.keys())):
            schema_violations.append(self._validate_one_required(h))
        schema_violations = pd.concat(schema_violations)
        print(f'There are {len(schema_violations)} violations of expected relationship number')
        if len(schema_violations):
            print(schema_violations)
        return duplicates, schema_violations

    def node_implies_plurality_of(self, a: str, b: str) -> Tuple[bool, TraversalPath, int, Type[Hierarchy]]:
        """
        returns: multiplicity, number, path
        """
        graph = self.relation_graph
        path = nx.shortest_path(graph, a, b)
        edges = [(x, y) for x, y in zip(path[:-1], path[1:])]
        multiplicity = [graph.edges[e]['multiplicity'] for e in edges]
        number = [graph.edges[e]['number'] for e in edges]
        direction = ['<-' if graph.edges[e].get('inverted', False) else '->' for e in edges]
        total_multiplicity = reduce(or_, multiplicity)
        total_number = sum(np.inf if i is None else i for i in number)
        if total_number == np.inf:
            total_number = None
        total_path = []
        for i, node in enumerate(path[1:]):
            total_path.append(direction[i])
            hier = self.singular_hierarchies[node]
            total_path.append(hier.__name__)
        return total_multiplicity, total_number, TraversalPath(*total_path), hier

    def is_factor_name(self, name):
        try:
            name = self.singular_name(name)
            return self.is_singular_factor(name) or self.is_singular_idname(name)
        except KeyError:
            return False

    def is_singular_idname(self, value):
        return value.split('.')[-1] in self.singular_idnames

    def is_plural_idname(self, value):
        return value.split('.')[-1] in self.plural_idnames

    def is_plural_factor(self, value):
        return value.split('.')[-1] in self.plural_factors

    def is_singular_factor(self, value):
        return value.split('.')[-1] in self.singular_factors

    def plural_name(self, singular_name):
        split = singular_name.split('.')
        before, singular_name = '.'.join(split[:-1]), split[-1]
        if singular_name in self.singular_idnames:
            return singular_name + 's'
        else:
            try:
                return before + self.singular_factors[singular_name] + 's'
            except KeyError:
                return before + self.singular_hierarchies[singular_name].plural_name

    def singular_name(self, plural_name):
        split = plural_name.split('.')
        before, plural_name = '.'.join(split[:-1]), split[-1]
        if self.is_singular_name(plural_name):
            return plural_name
        if plural_name in self.plural_idnames:
            return plural_name[:-1]
        else:
            try:
                return before + self.plural_factors[plural_name]
            except KeyError:
                return before + self.plural_hierarchies[plural_name].singular_name

    def is_plural_name(self, name):
        """
        Returns True if name is a plural name of a hierarchy
        e.g. spectra is plural for Spectrum
        """
        name = name.split('.')[-1]
        return name in self.plural_hierarchies or name in self.plural_factors or name in self.plural_idnames

    def is_singular_name(self, name):
        name = name.split('.')[-1]
        return name in self.singular_hierarchies or name in self.singular_factors or name in self.singular_idnames

    def __getitem__(self, address):
        return self.handler.begin_with_heterogeneous().__getitem__(address)

    def __getattr__(self, item):
        return self.handler.begin_with_heterogeneous().__getattr__(item)

    def plot_relations(self, show_hdus=True, fname='relations.pdf'):
        from networkx.drawing.nx_agraph import to_agraph
        if not show_hdus:
            G = nx.subgraph_view(self.relation_graph, lambda n: not self.relation_graph.nodes[n]['is_hdu'])
        else:
            G = self.relation_graph
        A = to_agraph(G)
        A.layout('dot')
        A.draw(fname)


from weaveio.utilities import quote


def property_value_in_hierarchy(name, **properties):
    blocks = []
    for property_name, property_value in properties.items():
        property_value = quote(property_value)
        blocks.append('\n'.join([f"//Begin WHERE {name} has {property_name}={property_value}",
        f"({name}.{property_name}={property_value} OR",
        f"    (",
        f"        EXISTS {{",
        f"            MATCH ({name})<-[:INDEXES]-(x)  // there is an index",
        f"            WHERE",
        f"               (EXISTS(x.{property_name}) AND x.{property_name}={property_value}) // index satisfies",
        f"            OR (EXISTS {{ (x)<-[:IS_REQUIRED_BY*]-(y {{{property_name}: {property_value}}}) }})  // index has it in hierarchy",
        f"            // original node has it in hierarchy",
        f"            OR (NOT EXISTS(x.{property_name}) AND NOT EXISTS {{ (x)<-[:IS_REQUIRED_BY*]-(y {{{property_name}: {property_value}}}) }}",
        f"                  AND EXISTS {{ ({name})<-[:IS_REQUIRED_BY*]-(parent {{{property_name}: {property_value}}}) }})",
        f"        }}",
        f"    )",
        f")"]))
    return '\nAND\n'.join(blocks)

def split_node_names(label, main_property, delimiter, *other_properties_to_set):
    s = f"""
MATCH (old: {label})
WHERE old.{main_property} CONTAINS '{delimiter}'
UNWIND split(old.{main_property}, "{delimiter}") as replacement
CALL apoc.refactor.cloneNodesWithRelationships([old])
YIELD input, output"""
    for p in other_properties_to_set+(main_property, ):
        s += f"\nSET output.{p} = replacement"
    return s + "\ndetach delete old"

def number_of_relationships(child_label, parent_label, lower, upper):
    s = f"""
match (children: {child_label})
unwind children as child
match (child)<-[]-(parent:{parent_label})
WITH child, parent, count(parent) as cnt"""
    wheres = []
    if lower is not None:
        wheres.append(f"cnt < {lower}")
    if upper is not None:
        wheres.append(f"cnt > {upper}")
    if len(wheres):
        s += '\nWHERE ' + ' OR '.join(wheres)
    s += f'\nRETURN "{parent_label}" as parent, "{child_label}" as child, child.id as child_id, ' \
         f'parent.id as parent_id, cnt'
    if lower is not None:
        s += f', {lower} as expected_lower'
    if upper is not None:
        s += f', {upper} as expected_upper'
    return s


import logging
import time
from collections import defaultdict
from copy import deepcopy
from typing import List

import networkx as nx
import numpy as np
import pandas as pd

from weaveio.address import Address
from weaveio.file import File
from weaveio.hierarchy import Graphable, Multiple
from weaveio.oldproduct import get_product
from weaveio.utilities import quote


class BasicQuery:
    def __init__(self, blocks: List = None, current_varname: str = None,
                 current_label: str = None, counter: defaultdict = None, termination_factor: str = None):
        self.blocks = [] if blocks is None else blocks
        self.current_varname = current_varname
        self.current_label = current_label
        self.counter = defaultdict(int) if counter is None else counter
        self.termination_factor = termination_factor

    def spawn(self, blocks, current_varname, current_label) -> 'BasicQuery':
        if self.termination_factor is not None:
            raise IndexError(f"Cannot continue query if it has terminated at a factor {self.termination_factor}")
        return BasicQuery(blocks, current_varname, current_label, self.counter)

    def terminate(self, blocks, current_varname, current_label, factor_name) -> 'BasicQuery':
        return BasicQuery(blocks, current_varname, current_label, self.counter, factor_name)

    def make(self, branch=False, orders_under=None) -> str:
        """
        Return Cypher query which will return json records with entries of HierarchyName and branch
        `HierarchyName` - The nodes to be realised
        branch - The complete branch for each `HierarchyName` node
        """
        if self.blocks:
            match = '\n\n'.join(['\n'.join(blocks) for blocks in self.blocks])
        else:
            raise ValueError(f"Cannot build a query with no MATCH statements")
        if self.termination_factor is not None:
            if branch:
                raise ValueError(f"May not return branches if returning a property value")
            returns = f"\nRETURN {self.current_varname}.id, " \
                      f"{self.current_varname}.{self.termination_factor}, " \
                      f"{{}} as branch, {{}} as indexer"
        else:
            if branch:
                returns = '\n'.join([f"OPTIONAL MATCH p1=({self.current_varname})<-[*]-(n1)",
                                     f"OPTIONAL MATCH p2=({self.current_varname})-[*]->(n2)",
                                     f"WITH collect(p2) as p2s, collect(p1) as p1s, {self.current_varname}",
                                     f"CALL apoc.convert.toTree(p1s+p2s) yield value",
                                     f"RETURN {self.current_varname} as {self.current_label}, value as branch, indexer"])
                returns = r'//Add Hierarchy Branch'+'\n' + returns
            else:
                returns = f"\nRETURN {self.current_varname}, {{}} as branch, indexer"
        indexers = f"\nOPTIONAL MATCH ({self.current_varname})<-[:INDEXES]-(indexer)"
        if orders_under is not None:
            filt_string = ', '.join([f"{k}: under.{k}" for k in orders_under])
            indexers += '\n' + '\n'.join([
                f"UNWIND $unders as under",
                f"MATCH p=({self.current_varname})<-[:IS_REQUIRED_BY*]-({{{filt_string}}})",
                "WITH *, relationships(p) as rels"])
            returns += ", [r IN rels | r.order] as orders"
        return f"{match}\n{indexers}\n{returns}"

    def index_by_address(self, address):
        if self.current_varname is None:
            name = '<first>'
        else:
            name = self.current_varname
        var_match = f"MATCH ({name}) "
        blocks = [[var_match]]
        for key, v in address.items():
            key = key.lower()
            count = self.counter[key]
            parent = f"{key}{count}"
            indexed_immediate = f"{key}{count+1}"
            indexed_far = f"{key}{count+2}"
            value = quote(v)
            path_match = f"OPTIONAL MATCH ({parent} {{{key}: {value}}})-[:IS_REQUIRED_BY*]->({name})"
            possible_index_match = f"OPTIONAL MATCH ({name})<-[:INDEXES]-({indexed_immediate})"
            possible_index_path_match = f"OPTIONAL MATCH ({indexed_immediate})<-[:IS_REQUIRED_BY*0..]-({indexed_far})"
            with_segment = f"WITH {parent}, {indexed_immediate}, {indexed_far}, {name}"
            where = '\n'.join(["WHERE",
                     f"    (EXISTS({name}.{key}) AND {name}.{key}={value}) OR",
                     f"    (EXISTS({indexed_immediate}.{key}) // immediate indexer must be {value}",
                     f"    AND {indexed_immediate}.{key}={value}) OR",
                     f"    (EXISTS({indexed_far}.{key}) // indexer must have parent {value}",
                     f"    AND {indexed_far}.{key}={value}) OR",
                     f"    (EXISTS({parent}.{key}) AND",
                     f"    NOT EXISTS({indexed_immediate}.{key}) AND",
                     f"    NOT EXISTS({indexed_far}.{key}) AND",
                     f"    {parent}.{key} = {value})"])
            block = [path_match, possible_index_match,possible_index_path_match, with_segment, where]
            self.counter[key] += 3
            blocks.append(block)
        return self.spawn(self.blocks + blocks, self.current_varname, self.current_label)

    def index_by_hierarchy_name(self, hierarchy_name, direction):
        name = '{}{}'.format(hierarchy_name.lower(), self.counter[hierarchy_name])
        if self.current_varname is None:
            first_encountered = False
            blocks = deepcopy(self.blocks)
            if len(blocks) == 0:
                blocks += [[f"MATCH ({name} :{hierarchy_name})"]]
            for i, block in enumerate(blocks):
                for j, line in enumerate(block):
                    if '<first>' in line and not first_encountered:
                        blocks[i][j] = line.replace('<first>', f'{name}:{hierarchy_name}')
                        first_encountered = True
                    else:
                        blocks[i][j] = line.replace('<first>', f'{name}')
            current_varname = name
        else:
            if direction == 'above':
                arrows = '<-[*]-'
            elif direction == 'below':
                arrows = '-[*]->'
            else:
                raise ValueError(f"Direction must be above or below")
            blocks = self.blocks + [[f"MATCH ({self.current_varname}){arrows}({name}:{hierarchy_name})"]]
            current_varname = name
        self.counter[hierarchy_name] += 1
        return self.spawn(blocks, current_varname, hierarchy_name)

    def index_by_id(self, id_value):
        blocks = self.blocks + [[f"WITH {self.current_varname}",
                                 f"WHERE {self.current_varname}.id = {quote(id_value)}"]]
        return self.spawn(blocks, self.current_varname, self.current_label)

    def select_factor_name(self, factor_name):
        return self.terminate(self.blocks, self.current_varname, self.current_label, factor_name)

    def select_factor_of(self, factor_name, hierarchy_name, direction):
        if hierarchy_name != self.current_label:
            hierarchy = self.index_by_hierarchy_name(hierarchy_name, direction)
            return hierarchy.select_factor_name(factor_name)
        else:
            return self.select_factor_name(factor_name)


class Indexable:
    def __init__(self, data, query: BasicQuery):
        self.data = data
        self.query = query

    def index_by_factor(self, factor_name):
        raise NotImplementedError(f"Please specify the parent hierarchy for {factor_name}")

    def index_by_single_hierarchy(self, hierarchy_name, direction):
        try:
            hierarchy = self.data.singular_hierarchies[hierarchy_name]
        except KeyError:
            return self.index_by_factor(hierarchy_name)
        name = hierarchy.__name__
        query = self.query.index_by_hierarchy_name(name, direction)
        return SingleHierarchy(self.data, query, hierarchy)

    def index_by_plural_hierarchy(self, hierarchy_name, direction):
        try:
            hierarchy = self.data.plural_hierarchies[hierarchy_name]
        except KeyError:
            return self.index_by_factor(hierarchy_name)
        query = self.query.index_by_hierarchy_name(hierarchy.__name__, direction)
        return HomogeneousHierarchy(self.data, query, hierarchy)

    def index_by_address(self, address):
        query = self.query.index_by_address(address)
        return HeterogeneousHierarchy(self.data, query)

    def implied_plurality_direction_of_node(self, name):
        """
        Returns True if the current hierarchy object expects name to be plural by looking at the
        relation graph
        """
        raise NotImplementedError


class HeterogeneousHierarchy(Indexable):
    """
	.<other> - data[Address(vph='green')].vph
		- if <other> is in the address, returns other
		- else raise IndexError
	.<other>s - data.OBs
		- returns HomogeneousStore
	[Address()] - data[Address(vph='green')]
		- Returns HeterogeneousStore filtered by the combined address
	[key]
		- Not implemented
    """
    def __getitem__(self, address):
        if isinstance(address, Address):
            return self.index_by_address(address)
        else:
            raise NotImplementedError("Cannot index by an id over multiple heterogeneous hierarchies")

    def implied_plurality_direction_of_node(self, name):
        return True, 'below'

    def index_by_hierarchy_name(self, hierarchy_name):
        if not self.data.is_plural_name(hierarchy_name):
            plural_name = self.data.plural_name(hierarchy_name)
            raise NotImplementedError(f"Can only get a single hierarchy when you have specified it. "
                                      f"Instead try `.{plural_name}`")
        return self.index_by_plural_hierarchy(hierarchy_name, 'below')

    def __getattr__(self, item):
        # if self.data.is_singular_idname(item) or self.data.is_plural_idname(item) or \
        #             self.data.is_singular_factor(item) or self.is_plural_factor(item):
        #     return self.index_by_factor(item)
        return self.index_by_hierarchy_name(item)


class Executable(Indexable):
    return_branch = False

    def __init__(self, data, query, nodetype):
        assert issubclass(nodetype, Graphable)
        super().__init__(data, query)
        self.nodetype = nodetype

    def __call__(self, payload=None):
        starts = time.perf_counter_ns(), time.process_time_ns()
        query_string = self.query.make(self.return_branch)
        result = self.data.graph.neograph.run(query_string, parameters=payload).to_table()  # type: py2neo.database.work.Table
        durations = (time.perf_counter_ns() - starts[0]) * 1e-9, (time.process_time_ns() - starts[1]) * 1e-9
        logging.info(f"Query completed in {durations[0]} secs ({durations[1]}) of which were process time")
        return self._process_result(result)

    def index_by_factor(self, factor_name):
        if self.data.is_plural_name(factor_name):
            name = self.data.singular_name(factor_name)
            plural_requested = True
        else:
            name = factor_name
            plural_requested = False
        plural, direction, path, number = self.data.node_implies_plurality_of(self.nodetype.singular_name, name)
        if plural and self.data.is_singular_name(factor_name):
            plural_name = self.data.plural_name(factor_name)
            raise KeyError(f"{self} has several possible {plural_name}. Please use `.{plural_name}` instead")
        nodetype = self.data.singular_hierarchies[path[-1]]
        query = self.query.select_factor_of(name, nodetype.__name__, direction)
        if plural or plural_requested:
            return HomogeneousFactor(self.data, query, nodetype, name)
        return SingleFactor(self.data, query, nodetype, name)

    def _process_result_row(self, row, nodetype):
        node, branch, indexer = row
        inputs = {}
        for f in nodetype.factors:
            inputs[f] = node[f]
        inputs[nodetype.idname] = node[nodetype.idname]
        try:
            base_query = self[node['id']]
        except TypeError:
            base_query = getattr(self, nodetype.plural_name)[node['id']]
        for p in nodetype.parents:
            if p.singular_name == nodetype.indexer:
                inputs[p.singular_name] = self._process_result_row([indexer, {}, {}], p)
            elif isinstance(p, Multiple):
                inputs[p.plural_name] = getattr(base_query, p.plural_name)
            else:
                inputs[p.singular_name] = getattr(base_query, p.singular_name)
        h = nodetype(**inputs)
        h.add_parent_query(base_query)
        return h

    def _process_result(self, result):
        if len(result) == 1 and result[0] is None:
            return []
        results = []
        for row in result:
            h = self._process_result_row(row, self.nodetype)
            results.append(h)
        return results


class ExecutableFactor(Executable):
    return_branch = False

    def __init__(self, data, query, nodetype, factor_name):
        super().__init__(data, query, nodetype)
        self.factor_name = factor_name


class SingleFactor(ExecutableFactor):
    def _process_result(self, result):
        return result[0][1]


class HomogeneousFactor(ExecutableFactor):
    def _process_result(self, result):
        if len(result) == 1 and result[0] is None:
            return []
        return [r[1] for r in result]

    def __iter__(self):
        ids = self()
        for i in ids:
            yield self.i


class ExecutableHierarchy(Executable):
    pass


class SingleHierarchy(ExecutableHierarchy):
    def __init__(self, data, query, nodetype, idvalue=None):
        super().__init__(data, query, nodetype)
        self.idvalue = idvalue

    def index_by_address(self, address):
        raise NotImplementedError("Cannot index a single hierarchy by an address")

    def index_by_hierarchy_name(self, hierarchy_name):
        if self.data.is_plural_name(hierarchy_name):
            multiplicity, direction, number = self.implied_plurality_direction_of_node(hierarchy_name)
            return self.index_by_plural_hierarchy(hierarchy_name, direction)
        elif self.data.is_singular_name(hierarchy_name):
            multiplicity, direction, number = self.implied_plurality_direction_of_node(hierarchy_name)
            if multiplicity:
                plural = self.data.plural_name(hierarchy_name)
                raise KeyError(f"{self} has several possible {plural}. Please use `.{plural}` instead")
            return self.index_by_single_hierarchy(hierarchy_name, direction)
        else:
            raise KeyError(f"{hierarchy_name} is an unknown factor/hierarchy")

    def implied_plurality_direction_of_node(self, name):
        if self.data.is_plural_name(name):
            name = self.data.singular_name(name)
        if not self.data.is_singular_name(name):
            raise ValueError(f"{name} is not recognised as a factor/hierarchy name")
        return self.data.node_implies_plurality_of(self.nodetype.singular_name, name)[:-1]

    def __getattr__(self, item):
        if item in getattr(self.nodetype, 'products', []):
            return Products(self, item)
        return self.index_by_hierarchy_name(item)

    def __call__(self):
        rs = super().__call__()
        assert len(rs) == 1
        return rs[0]


class HomogeneousHierarchy(ExecutableHierarchy):
    """
	.<other> - OBs.OBspec
		- returns Hierarchy/factor/id/file
	.<other>s
		- return HomogeneousStore (e.g. ob.l1.singles)
	[Address()]
		- return Hierarchy if address describes a unique object
		- return HomogeneousStore if address contains some missing factors
		- return HomogeneousStore  if not
		- raise IndexError if address is incompatible
	[key] - OBs[obid]
		- if indexable by key, return Hierarchy
    """
    def index_by_id(self, idvalue):
        query = self.query.index_by_id(idvalue)
        return SingleHierarchy(self.data, query, self.nodetype, idvalue)

    def implied_plurality_direction_of_node(self, name):
        return self.data.node_implies_plurality_of(self.nodetype.singular_name, name)

    def index_by_address(self, address):
        query = self.query.index_by_address(address)
        return HomogeneousHierarchy(self.data, query, self.nodetype)

    def __getitem__(self, item):
        if isinstance(item, Address):
            return self.index_by_address(item)
        else:
            return self.index_by_id(item)

    def __getattr__(self, item):
        if item in getattr(self.nodetype, 'products', []):
            return Products(self, item)
        if self.data.is_plural_name(item):
            name = self.data.singular_name(item)
        else:
            raise ValueError(f"{self} requires a plural {item}, try `.{self.data.plural_name(item)}`")
        _, direction, _ = self.implied_plurality_direction_of_node(name)
        return self.index_by_plural_hierarchy(item, direction)

    def __iter__(self):
        ids = getattr(self, self.nodetype.idname+'s')()
        for i in ids:
            yield self[i]


class Products:
    def __init__(self, filenode, product_name, index=None):
        self.filenode = filenode
        self.query = filenode.query
        self.product_name = product_name
        assert issubclass(filenode.nodetype, File)
        self.default_key = self.filenode.nodetype.product_indexables[product_name]
        graph = self.filenode.data.relation_graph  # type: nx.DiGraph
        if self.default_key is None:
            self.available_keys = None
        else:
            self.available_keys = [f for h in nx.ancestors(graph, filenode.nodetype.singular_name)
                                   for f in graph.nodes[h]['factors']]
        if self.default_key is None and index is not None:
            raise ValueError(f"{filenode.nodetype.singular_name}.{product_name} cannot be indexed")
        if index is not None:
            if isinstance(index, (list, tuple, np.ndarray)):
                index = {self.default_key: index}
            elif isinstance(index, dict):
                index = {k: v if isinstance(v, (list, tuple, np.ndarray)) else v for k, v in index.items()}
            else:
                index = {self.default_key: index}
        self.index = index

    def __getitem__(self, item):
        if isinstance(item, Address):
            return Products(self.filenode.__getitem__(item), self.product_name, self.index)
        return Products(self.filenode, self.product_name, item)

    def __getattr__(self, item):
        if item == self.filenode.nodetype.singular_name:
            return self.filenode
        return self.filenode.__getattr__(item)

    def _process_result(self, result):
        if len(result) == 0:
            return []
        l = np.asarray(result['orders'].tolist())
        which = np.sum(l != 0, axis=0)
        if sum(which > 0) > 1:
            raise NotImplementedError(f"Cannot get data from a multi-index search, please report this")
        try:
            indices = l[:, np.where(which>0)[0][0]]
        except IndexError:
            indices = l[:, 0]  # they are all 0
        product_type = self.filenode.nodetype.products[self.product_name]
        data = []
        for fname, index in zip(result.iloc[:, 1], indices):
            datum = getattr(self.filenode.nodetype(fname), f'read_{self.product_name}')().data
            if self.index is not None:
                datum = datum[index]
            data.append(product_type(datum, None))
        if len(data) == 1:
            return data[0].data
        return product_type.concatenate_data(*data)

    def __iter__(self):
        ids = getattr(self.filenode, self.default_key+'s')()
        for i in ids:
            yield self[i]

    def __call__(self):
        under_type = self.filenode.nodetype.product_indexables[self.product_name]
        starts = time.perf_counter_ns(), time.process_time_ns()
        if self.index is None:
            query_string = self.filenode.fnames.query.make(False)
            result = self.filenode.data.graph.neograph.run(query_string).to_data_frame()
            result['orders'] = 0
            result['orders'] = result['orders'].apply(lambda x: [x])  # just to get a list
        else:
            orders_under = list(self.index.keys())
            query_string = self.filenode.fnames.query.make(False, orders_under)
            index = []
            for i in range(max(len(vs) for vs in self.index.values())):
                d = {}
                for k, v in self.index.items():
                    try:
                        d[k] = v[i]
                    except IndexError:
                        d[k] = v
                index.append(d)
            result = self.filenode.data.graph.neograph.run(query_string, parameters={'unders': index}).to_data_frame()
        durations = (time.perf_counter_ns() - starts[0]) * 1e-9, (time.process_time_ns() - starts[1]) * 1e-9
        logging.info(f"Query completed in {durations[0]} secs ({durations[1]}) of which were process time")
        return self._process_result(result)



from functools import reduce
from typing import Union, List, Dict, Any, Type, Iterable
from pathlib import Path

from astropy.io import fits
import numpy as np
import pandas as pd
from astropy.table import vstack, Table as AstropyTable


def header2table(header):
    keys = set(header.keys())
    data = AstropyTable([[header[k]] for k in keys], names=list(keys))
    return data


class Product:
    def __init__(self, data, index):
        self.data = data
        self.index = index

    @classmethod
    def concatenate(cls, *products):
        return cls(cls.concatenate_data(*products), cls.concatenate_index(*products))

    @classmethod
    def concatenate_index(cls, *products):
        return  pd.concat([p.index for p in products])

    @classmethod
    def concatenate_data(cls, *products):
        raise NotImplementedError

    def sort(self):
        raise NotImplementedError

    def __getitem__(self, item: pd.Series) -> 'Product':
        if isinstance(item, pd.Series):
            item = item.values
        return self.__class__(self.data[item], self.index[item])


class Array(Product):
    def __init__(self, data, index):
        super().__init__(np.asarray(data), index)

    @classmethod
    def concatenate_data(cls, *products):
        return np.stack([p.data for p in products])


class Table(Product):
    def __init__(self, data, index):
        super().__init__(AstropyTable(data), index)

    @classmethod
    def concatenate_data(cls, *products):
        return vstack([p.data for p in products])


class Header(Product):
    @classmethod
    def concatenate_data(cls, *products):
        return vstack([header2table(p.data) for p in products])


def get_product(files: List['File'], product_name: str, product_index: pd.DataFrame = None) -> Product:
    product_type = files[0].products[product_name]
    for f in files:
        if f.products[product_name] is not product_type:
            raise TypeError(f"Stacking the {product_name} product from this query is not supported since {product_name} implies different types of products")
        if not f.is_concatenatable_with(files[0], product_name):
            raise TypeError(f"Stacking the {product_name} product from this query is not supported since they are unequal in concatenation_constants")
    products = []
    for f in files:
        product = f.read_product(product_name)
        if product_index is not None:
            filt = f.match_index(product_index)
            product = product[filt]
        products.append(product)
    return product_type.concatenate(*products).data


from .writequery import CypherData, unwind, collect, groupby


def get_fibretargets(df_svryinfo, df_fibinfo):
    srvyinfo = CypherData(df_svryinfo)   # surveys split inline
    fibinfo = CypherData(df_fibinfo)
    with unwind(srvyinfo) as svryrow:
        with unwind(svryrow['surveyname']) as surveyname:
            survey = Survey(surveyname=surveyname)
        surveys = collect(survey)
        prog = SubProgramme(targprog=svryrow['targprog'], surveys=surveys)
        cat = SurveyCatalogue(catname=svryrow['targcat'], subprogramme=prog)
    cat_collection = collect(cat)
    cats = groupby(cat_collection, 'targcat')
    with unwind(fibinfo) as fibrow:
        cat = cats[fibrow['targcat']]
        weavetarget = WeaveTarget(cname=fibrow['cname'])
        surveytarget = SurveyTarget(surveycatalogue=cat, weavetarget=weavetarget, tables=fibrow)
        fibre = Fibre(fibreid=fibrow['fibreid'])
        fibtarget = FibreTarget(surveytarget=surveytarget, fibre=fibre)
    return collect(fibtarget)


def add_raw():
    fibretargets = get_fibretargets()
    obspec = OBSpec(xml=xml, obtitle=obtitle, obstemp=obstemp, progtemp=progtemp)
    obspec.fibreset.attach(fibretargets=fibretargets)

    ob = OB(obid=obid, obstartmjd=mjd, obspec=obspec)
    exposure = Exposure(expmjd=expmjd, ob=ob)
    run = Run(runid=runid, exposure=exposure, armconfig=armconfig)
    system = System(sysver)
    simulator = Simulator(simver)
    casu = CASU(casuver)
    raw = RawSpectrum(run=run, casu=casu, simulator=simulator, system=system)
    rawfile = RawFile(fname=fname, raw=raw)


def add_single():
    system = System(sysver)
    simulator = Simulator(simver)
    casu = CASU(casuver)
    run = Run(runid=runid, exposure=None)  # implies matching not creation
    raw = RawSpectrum(run=run, casu=casu, simulator=simulator, system=system)
    fibretargets = get_fibretargets()
    with unwind(fibretargets, checksums) as (fibretarget, checksum):
        spec = L1SingleSpectrum(rawspectrum=raw, fibretarget=fibretarget, casu=casu, checksum=checksum)
    specs = collect(spec)
    single = L1SingleFile(fname=fname, l1singlespectra=specs)


def add_stack():
    system = System(sysver)
    simulator = Simulator(simver)
    casu = CASU(casuver)
    fibretargets = get_fibretargets()
    with unwind(checksums, fibretargets) as (checksum, fibretarget):
        with unwind(runids) as runid:
            run = Run(runid=runid, exposure=None)  # implies matching not creation
            raw = RawSpectrum(run=run, casu=casu, simulator=simulator, system=system)
            spec = L1SingleSpectrum(rawspectrum=raw, fibretarget=fibretarget, casu=casu, checksum=None)
        specs = collect(spec)
        stack = L1StackSpectrum(l1singlespectra=specs, casu=casu, checksum=checksum)
    stacks = collect(stack)
    L1StackFile(fname=fname, l1stackspectra=stacks)





from pathlib import Path

import numpy as np
from astropy.table import Table
from astropy.io import fits
from astropy import units as u, constants
import matplotlib.pyplot as plt


class BasicSpectra:
    ulam = u.AA
    uadu = u.adu

    def __init__(self, wvls, adu, aduivar, calibration):
        self.ucal = calibration.unit
        self.uivaradu = 1 / self.uadu / self.uadu
        self.uflam = self.ucal * u.adu
        self.uivarflam = 1 / (self.uflam ** 2)
        self._wvls = wvls.to(self.ulam)
        self._adu = np.atleast_2d(adu).to(self.uadu)
        self._aduivar = np.atleast_2d(aduivar).to(self.uivaradu)
        self._calibration = np.atleast_2d(calibration).to(self.ucal)
        if self._calibration.shape != self._aduivar.shape != self._adu.shape:
            raise TypeError("Shapes do not match")
        self._flux = (self._adu * self._calibration).to(self.uflam)
        self._fluxivar = (self._aduivar / (self._calibration ** 2)).to(self.uivarflam)
        self._fluxerr = 1 / np.sqrt(self._fluxivar)
        self._aduerr = 1 / np.sqrt(self._aduivar)
        self._freqs = (constants.c / wvls).to(u.Hz)

    @property
    def wvls(self):
        return self._wvls

    @property
    def adu(self):
        return self._adu

    @property
    def aduivar(self):
        return self._aduivar

    @property
    def calibration(self):
        return self._calibration

    @property
    def flux(self):
        return self._flux

    @property
    def fluxivar(self):
        return self._fluxivar

    @property
    def fluxerr(self):
        return self._fluxerr

    @property
    def aduerr(self):
        return self._aduerr

    def __len__(self):
        return self.flux.shape[0]

    def __repr__(self):
        return f"<{self.__class__.__name__}({len(self)} in {self.uflam})"

    @property
    def snr(self):
        return (self.adu / self.aduerr).value

    def plot(self, ax=None, fluxtype='flux', xunit='lam', legend=False, **kwargs):
        return self[:].plot(ax, fluxtype, xunit, legend, **kwargs)

    @classmethod
    def from_weave_obfits(cls, hdus, flux_hdu: int, ivar_hdu: int, calib_hdu: int):
        wvl_start = hdus[flux_hdu].header['CRVAL1']
        wvl_inc = hdus[flux_hdu].header['CD1_1']
        flux = hdus[flux_hdu].data
        invflux = hdus[ivar_hdu].data
        calibration = hdus[calib_hdu].data
        wvl_end = (flux.shape[1] * wvl_inc) + wvl_start
        wvls = np.arange(wvl_start, wvl_end, wvl_inc) * u.AA
        cal = calibration * u.erg / u.s / u.cm / u.cm / u.adu / u.AA
        return cls(wvls, flux * u.adu, invflux / u.adu / u.adu, cal)



class BasicIndexedSpectra(BasicSpectra):
    def __init__(self, host_spectra, index):
        self.host_spectra = host_spectra
        self.index = index

    @property
    def uadu(self):
        return self.host_spectra.uadu
    @property
    def uflam(self):
        return self.host_spectra.uflam
    @property
    def ucal(self):
        return self.host_spectra.ucal
    @property
    def uwvl(self):
        return self.host_spectra.uwvl
    @property
    def uivaradu(self):
        return self.host_spectra.uivaradu
    @property
    def uivarflam(self):
        return self.host_spectra.uivarflam

    @property
    def wvls(self):
        return self.host_spectra.wvls

    @property
    def freqs(self):
        return self.host_spectra.freqs

    @property
    def names(self):
        return self.host_spectra.names[self.index]

    @property
    def names_array(self):
        return self.host_spectra.names_array[self.index]

    @property
    def adu(self):
        return self.host_spectra.adu[self.index]

    @property
    def aduivar(self):
        return self.host_spectra.aduivar[self.index]

    @property
    def calibration(self):
        return self.host_spectra.calibration[self.index]

    @property
    def flux(self):
        return self.host_spectra.flux[self.index]

    @property
    def fluxivar(self):
        return self.host_spectra.fluxivar[self.index]

    @property
    def fluxerr(self):
        return self.host_spectra.fluxerr[self.index]

    @property
    def aduerr(self):
        return self.host_spectra.aduerr[self.index]

    @property
    def arms(self):
        return list(map(str, np.asarray(self.host_spectra.arms)[self.index]))

    @property
    def resolutions(self):
        return list(map(str, np.asarray(self.host_spectra.resolutions)[self.index]))

    def squeeze(self):
        if len(self) == 1:
            return BasicSpectrum(self)
        else:
            return self

    def plot(self, ax=None, fluxtype='flux', xunit='lam', legend=False, **kwargs):
        if fluxtype == 'flux':
            flux = self.flux.to(self.uflam)
        elif fluxtype == 'counts' or fluxtype == 'adu':
            flux = self.adu
        else:
            raise ValueError(f"{fluxtype} is not a valid fluxtype")
        if xunit == 'lam':
            x = self.wvls
            xx = fr"$\lambda / {self.wvls.unit.to_string('latex').strip('$')}$"
        elif xunit == 'freq':
            x = self.freqs
            xx = fr"$\nu / {self.freqs.unit.to_string('latex').strip('$')}$"
        elif xunit == 'loglam':
            x = np.log10(self.wvls.value)
            xx = fr"$\log[\lambda / {self.wvls.unit.to_string('latex').strip('$')}]$"
        elif xunit == 'logfreq':
            x = np.log10(self.freqs.value)
            xx = fr"$\log[\nu / {self.freqs.unit.to_string('latex').strip('$')}]$"
        else:
            raise ValueError(f"{xunit} is not valid")
        if ax is None:
            fig, ax = plt.subplots()
        if flux.ndim == 2:
            for f, n in zip(flux, self.names):
                ax.plot(x, f, label=n, **kwargs)
        else:
            ax.plot(x, flux, label=self.names, **kwargs)
        ax.set_ylabel(f"{fluxtype} / {flux.unit.to_string('latex')}")
        ax.set_xlabel(xx)
        if legend:
            plt.legend()
        return ax


class BasicSpectrum(BasicIndexedSpectra):
    def __init__(self, host_spectra):
        super().__init__(host_spectra, 0)

    def __repr__(self):
        return f"<Spectrum({self.names} / {self.uflam})>"


class HierarchicalMixin:
    def __init__(self, wvls, adu, invadu, calibration, hierarchy):
        self.hierarchy = hierarchy
        names = hierarchy.runids
        groupname = hierarchy.name
        super(HierarchicalMixin, self).__init__(wvls, adu, invadu, calibration, names, groupname)

    @property
    def address(self):
        return self.hierarchy.address

    @classmethod
    def from_hierarchy(cls, hierarchy):
        raise NotImplementedError


class Spectra(HierarchicalMixin, BasicSpectra):
    pass


class Spectrum(HierarchicalMixin, BasicSpectrum):
    pass


from collections import defaultdict
from typing import List, Union, Any, Type, Tuple

import py2neo

from .common import FrozenQuery, AmbiguousPathError
from .factor import SingleFactorFrozenQuery, ColumnFactorFrozenQuery, RowFactorFrozenQuery, TableFactorFrozenQuery
from .tree import Branch, TraversalPath
from ..hierarchy import Hierarchy, Multiple
from ..utilities import quote
from ..writequery import CypherVariable


class HierarchyFrozenQuery(FrozenQuery):
    def __getitem__(self, item):
        raise NotImplementedError

    def __getattr__(self, item):
        raise NotImplementedError


class HeterogeneousHierarchyFrozenQuery(HierarchyFrozenQuery):
    """The start point for building queries"""
    executable = False

    def __repr__(self):
        return f'query("{self.data.rootdir}/")'

    def __getattr__(self, item):
        if item in self.data.plural_factors:
            return self._get_plural_factor(item)
        elif item in self.data.singular_factors:
            raise AmbiguousPathError(f"Cannot return a single factor from a heterogeneous dataset")
        elif item in self.data.singular_hierarchies:
            raise AmbiguousPathError(f"Cannot return a singular hierarchy without filtering first")
        else:
            name = self.data.singular_name(item)
            return self._get_plural_hierarchy(name)

    def _get_plural_hierarchy(self, hierarchy_name) -> 'HomogeneousHierarchyFrozenQuery':
        hier = self.data.singular_hierarchies[hierarchy_name]
        new = self.branch.handler.begin(hier.__name__)
        return HomogeneousHierarchyFrozenQuery(self.handler, new, hier, new.current_hierarchy, self)

    def _get_plural_factor(self, factor_name):
        hierarchy_name, factor_name, singular_name = self.handler.hierarchy_of_factor(factor_name)
        return self._get_plural_hierarchy(hierarchy_name)._get_plural_factor(factor_name)


class DefiniteHierarchyFrozenQuery(HierarchyFrozenQuery):
    """The template class for hierarchy classes that are not heterogeneous i.e. they have a defined hierarchy type"""
    SingleFactorReturnType = None

    def __init__(self, handler, branch: Branch, hierarchy_type: Type[Hierarchy], hierarchy_variable: CypherVariable, parent: 'FrozenQuery'):
        super().__init__(handler, branch, parent)
        self.hierarchy_type = hierarchy_type
        self.hierarchy_variable = hierarchy_variable

    def _prepare_query(self):
        """Add a hierarchy node return statement"""
        query = super(DefiniteHierarchyFrozenQuery, self)._prepare_query()
        with query:
            query.returns(self.branch.find_hierarchies()[-1])
        return query

    def _process_result_row(self, row, nodetype):
        node = row[0]
        inputs = {}
        for f in nodetype.factors:
            inputs[f] = node[f]
        inputs[nodetype.idname] = node[nodetype.idname]
        base_query = getattr(self.handler.begin_with_heterogeneous(), nodetype.plural_name)[node['id']]
        for p in nodetype.parents:
            if isinstance(p, Multiple):
                inputs[p.plural_name] = getattr(base_query, p.plural_name)
            else:
                try:
                    inputs[p.singular_name] = getattr(base_query, p.singular_name)
                except AmbiguousPathError:
                    inputs[p.singular_name] = getattr(base_query, p.plural_name)  # this should not have to be done
        h = nodetype(**inputs, do_not_create=True)
        h.add_parent_query(base_query)
        return h

    def _post_process(self, result: py2neo.Cursor):
        result = result.to_table()
        if len(result) == 1 and result[0] is None:
            return []
        results = []
        for row in result:
            h = self._process_result_row(row, self.hierarchy_type)
            results.append(h)
        return results

    def node_implies_plurality_of(self, end: str) -> Tuple[bool, TraversalPath, int, Type[Hierarchy]]:
        if end in self.data.singular_factors or end in self.data.singular_idnames:
            hier_name = self.handler.hierarchy_of_factor(end)[0]
        elif end in self.data.plural_factors or end in self.data.plural_idnames:
            hier_name = self.handler.hierarchy_of_factor(self.data.singular_name(end))[0]
        else:
            hier_name = end
        start = self.hierarchy_type.singular_name.lower()
        if start == hier_name:
            return False, None, 1, self.hierarchy_type
        multiplicity, number, path, hier = self.data.node_implies_plurality_of(start, hier_name)
        return multiplicity, path, number, hier

    def _get_plural_hierarchy(self, name):
        multiplicity, path, number, hier = self.node_implies_plurality_of(name)
        # dont check for multiplicity here, since plural is requested anyway
        new = self.branch.traverse(path)
        return HomogeneousHierarchyFrozenQuery(self.handler, new, hier, new.current_hierarchy, self)

    def _get_factor_query(self, *names) -> Tuple[Branch, List[CypherVariable], List[bool], List[int]]:
        multiplicities = []
        numbers = []
        factor_variables = []
        branch = self.branch
        for name in names:
            hierarchy_name, factor_name, singular_name = self.handler.hierarchy_of_factor(name, self.hierarchy_type)
            friendly_name = '_'.join(name.split('.'))
            if hierarchy_name != self.hierarchy_type.singular_name:  # if it refers to another hierarchy
                multiplicity, path, number, _ = self.node_implies_plurality_of(hierarchy_name)
                branch = branch.traverse(path)
            else:
                multiplicity = False
                number = 1
                if not (singular_name in self.hierarchy_type.factors or singular_name == self.hierarchy_type.idname):
                    raise KeyError(f"{self} does not have factor {singular_name}")
            operated = branch.operate('{h}.' + f'{singular_name} as {friendly_name}', h=self.hierarchy_variable)
            if multiplicity:
                new = branch.collect([], [operated])  # collapse back to the original level
                factor_variable = new.action.transformed_variables[operated.current_variables[0]]
            else:
                new = operated
                factor_variable = new.current_variables[0]
            multiplicities.append(multiplicity)
            numbers.append(number)
            factor_variables.append(factor_variable)
        return branch, factor_variables, multiplicities, numbers

    def _get_plural_factor(self, name):
        singular_name = self.data.singular_name(name)
        branch, factor_variables, multiplicities, numbers = self._get_factor_query(singular_name)
        if self.data.is_singular_name(name) and multiplicities[0]:
            plural = self.data.plural_name(singular_name)
            raise AmbiguousPathError(f"{self} has multiple {plural}, you need to explicitly pluralise them.")
        return ColumnFactorFrozenQuery(self.handler, branch, [name], factor_variables, numbers, self)

    def _get_factor_table_query(self, item):
        """
        __getitem__ is for returning factors and ids
        There are three types of getitem input values:
        List: [[a, b]], where labelled table-like rows are output
        Tuple: [a, b], where a list of unlabelled dictionaries are output
        str: [a], where a single value is returned

        In all three cases, you still need to specify plural or singular forms.
        This allows you to have a row of n dimensional heterogeneous data.
        returns query and the labels (if any) for the table
        """
        if isinstance(item, tuple):  # return without headers
            return_keys = None
            keys = list(item)
        elif isinstance(item, list):
            keys = item
            return_keys = item
        elif item is None:
            raise TypeError("item must be of type list, tuple, or str")
        else:
            raise KeyError(f"Unknown item {item} for `{self}`")
        branch, factor_variables, multiplicities, numbers = self._get_factor_query(*keys)
        expected_multi = [k for m, k in zip(multiplicities, keys) if self.data.is_singular_name(k) and m]
        if expected_multi:
            plurals = [self.data.plural_name(i) for i in expected_multi]
            raise AmbiguousPathError(f"Each {self.hierarchy_type} in {self} has multiple `{', '.join(plurals)}`, you need to explicitly pluralise them.")
        return branch, return_keys, factor_variables, numbers

    def _get_single_factor_query(self, item):
        branch, factor_variables, multiplicities, numbers = self._get_factor_query(item)
        if multiplicities[0] and self.data.is_singular_name(item):
            plural = self.data.plural_name(item)
            raise AmbiguousPathError(f"Each `{self.hierarchy_type.singular_name}` in `{self}` has multiple `{plural}`, you need to explicitly use `{plural}`.")
        return self.SingleFactorReturnType(self.handler, branch, [item], factor_variables, numbers, self)

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._get_single_factor_query(item)
        return self._get_factor_table_query(item)

    def __getattr__(self, item):
        if self.data.is_plural_name(item) and self.data.is_factor_name(item):
            return self._get_plural_factor(item)
        elif item in self.data.singular_hierarchies:
            return self._get_singular_hierarchy(item)
        elif item in self.data.plural_hierarchies:
            name = self.data.singular_name(item)
            return self._get_plural_hierarchy(name)
        else:
            raise AttributeError(f"{self} has no attribute {item}")


class SingleHierarchyFrozenQuery(DefiniteHierarchyFrozenQuery):
    """Contains only a single hierarchy type with an identifier"""
    SingleFactorReturnType = SingleFactorFrozenQuery

    def __init__(self, handler, branch: Branch, hierarchy_type: Type[Hierarchy], hierarchy_variable: CypherVariable,  identifier: Any, parent: 'FrozenQuery'):
        super().__init__(handler, branch, hierarchy_type, hierarchy_variable, parent)
        self._identifier = identifier

    def __getattr__(self, item):
        if self.data.is_singular_name(item) and self.data.is_factor_name(item):
            return self._get_singular_factor(item)
        return super().__getattr__(item)

    def _get_factor_table_query(self, keys):
        branch, return_keys, factor_variables, numbers = super()._get_factor_table_query(keys)
        return RowFactorFrozenQuery(self.handler, branch, keys, factor_variables, numbers, return_keys, self)

    def __repr__(self):
        if self._identifier is None:
            return f'{self.parent}.{self.hierarchy_type.singular_name}'
        return f'{self.parent}[{quote(self._identifier)}]'

    def _get_singular_hierarchy(self, name):
        multiplicity, path, number, hier = self.node_implies_plurality_of(name)
        if multiplicity:
            plural = self.data.plural_name(name)
            raise AmbiguousPathError(f"You have requested a single {name} but {self} has multiple {plural}. Use .{plural}")
        branch = self.branch.traverse(path)
        return SingleHierarchyFrozenQuery(self.handler, branch, hier, branch.current_hierarchy, None, self)

    def _get_singular_factor(self, name):
        branch, factor_variables, multiplicities, numbers = self._get_factor_query(name)
        if multiplicities[0]:
            plural = self.data.plural_name(name)
            raise AmbiguousPathError(f"{self} has multiple {name}s. Use {plural} instead")
        return SingleFactorFrozenQuery(self.handler, branch, name, factor_variables, numbers, self)

    def _post_process(self, result: py2neo.Cursor):
        rows = super()._post_process(result)
        if len(rows) != 1:
            idents = defaultdict(list)
            for frozen in self._traverse_frozenquery_stages():
                if isinstance(frozen, SingleHierarchyFrozenQuery):
                    idents[frozen.hierarchy_type.idname].append(frozen._identifier)
                elif isinstance(frozen, IdentifiedHomogeneousHierarchyFrozenQuery):
                    idents[frozen.hierarchy_type.idname] += frozen._identifiers
            if idents:
                d = {k: [i for i in v if i is not None] for k, v in idents.items()}
                d = {k: v for k, v in d.items() if len(v)}
                raise KeyError(f"One or more identifiers in {d} are not present in the database")
        return rows[0]


class HomogeneousHierarchyFrozenQuery(DefiniteHierarchyFrozenQuery):
    """A list of hierarchies of the same type that are not identified"""
    SingleFactorReturnType = ColumnFactorFrozenQuery

    def __repr__(self):
        return f'{self.parent}.{self.hierarchy_type.plural_name}'

    def _get_factor_table_query(self, item):
        branch, return_keys, factor_variables, numbers = super()._get_factor_table_query(item)
        return TableFactorFrozenQuery(self.handler, branch, item, factor_variables, numbers, return_keys, self)

    def __getitem__(self, item):
        """
        Returns a table of factor values or (if that fails) a filter by identifiers
        """
        try:
            return super(HomogeneousHierarchyFrozenQuery, self).__getitem__(item)
        except KeyError:
            if isinstance(item, (list, tuple)):
                disallowed_factors = [i for i in item if self.data.is_factor_name(i)]
                if disallowed_factors:
                    ids = list(set(item) - set(disallowed_factors))
                    raise SyntaxError(f"You cannot index factors and hierarchies at the same time. "
                                      f"Separate your queries for {ids} and `{disallowed_factors}`")
                return self._filter_by_identifiers(item)
            return self._filter_by_identifier(item)

    def __getattr__(self, item):
        if item in self.data.singular_hierarchies:
            multiplicity, path, number, hier = self.node_implies_plurality_of(item)
            if multiplicity:
                plural = self.data.plural_name(item)
                raise AmbiguousPathError(f"{self} has multiple {plural}. Use .{plural} instead")
        if self.data.is_singular_name(item) and self.data.is_factor_name(item):
            hierarchy_name, factor_name, singular_name = self.handler.hierarchy_of_factor(item)
            multiplicity, path, number, hier = self.node_implies_plurality_of(singular_name)
            if multiplicity:
                plural = self.data.plural_name(item)
                raise AmbiguousPathError(f"{self} has multiple {plural}. Use .{plural} instead.")
        return super(HomogeneousHierarchyFrozenQuery, self).__getattr__(item)

    def _filter_by_identifiers(self, identifiers: List[Union[str, int, float]]) -> 'IdentifiedHomogeneousHierarchyFrozenQuery':
        idname = self.hierarchy_type.idname
        new = self.branch.add_data(identifiers)
        identifiers_var = new.current_variables[0]
        branch = new.filter('{h}.' + idname + ' in {identifiers}', h=self.hierarchy_variable, identifiers=identifiers_var)
        return IdentifiedHomogeneousHierarchyFrozenQuery(self.handler, branch, self.hierarchy_type, self.hierarchy_variable, identifiers, self)

    def _filter_by_identifier(self, identifier: Union[str, int, float]):
        idname = self.hierarchy_type.idname
        new = self.branch.add_data(identifier)
        identifier_var = new.current_variables[0]
        branch = new.filter('{h}.' + idname + ' = {identifier}', h=self.hierarchy_variable, identifier=identifier_var)
        if isinstance(self.parent, (HeterogeneousHierarchyFrozenQuery, SingleHierarchyFrozenQuery)):
            return SingleHierarchyFrozenQuery(self.handler, branch, self.hierarchy_type, self.hierarchy_variable, identifier, self)
        else:
            raise AmbiguousPathError(f"`{self.parent}` is plural, to identify `{self}` by id, you must use "
                                     f"`{self}[[{quote(identifier)}]]` instead of "
                                     f"`{self}[{quote(identifier)}]`.")


class IdentifiedHomogeneousHierarchyFrozenQuery(HomogeneousHierarchyFrozenQuery):
    """
    An ordered duplicated list of hierarchies each identified by an id
    If an id appears more than once, it will be duplicated appropriately
    The list is ordered by id input order
    """
    def __init__(self, handler, branch: Branch, hierarchy_type: Type[Hierarchy], hierarchy_variable: CypherVariable, identifiers: List[Any], parent: 'FrozenQuery'):
        super().__init__(handler, branch, hierarchy_type, hierarchy_variable, parent)
        self._identifiers = identifiers

    def __repr__(self):
        return f'{self.parent}.{self.hierarchy_type.plural_name}[{self._identifiers}]'

    def _post_process(self, result: py2neo.Cursor):
        r = super(IdentifiedHomogeneousHierarchyFrozenQuery, self)._post_process(result)
        ids = set(i.identifier for i in r)
        missing = [i for i in self._identifiers if i not in ids]
        if any(missing):
            raise KeyError(f"{self.hierarchy_type.idname} {missing} not found")
        return r



from typing import Tuple

import networkx as nx

from .common import AmbiguousPathError
from .hierarchy import *
from .tree import BranchHandler


class Handler:
    def __init__(self, data):
        self.data = data
        self.branch_handler = data.branch_handler

    def begin_with_heterogeneous(self):
        return HeterogeneousHierarchyFrozenQuery(self, self.branch_handler.entry)

    def hierarchy_of_factor(self, factor_name: str, start: Type[Hierarchy] = None) -> Tuple[str, str, str]:
        """
        returns the hierarchy_name, factor_name, singular_name
        """
        namelist = factor_name.split('.')
        hierarchy_names = None
        singular_name = None
        if len(namelist) > 1:
            hierarchy_name_list, factor_name = namelist[:-1], namelist[-1]
            singular_name = self.data.singular_name(factor_name)
            if not all(h in self.data.singular_hierarchies for h in hierarchy_name_list):
                raise KeyError(f"Not all of {hierarchy_name_list} are valid factor names")
            if not nx.has_path(self.data.relation_graph, hierarchy_name_list[0], hierarchy_name_list[-1]):
                raise KeyError(f"The path {'-'.join(hierarchy_name_list)} does not exist in the schema")
            hierarchy_names = []
            for n in self.data.factor_hierarchies[singular_name]:
                if hierarchy_name_list[-1] == n or nx.has_path(self.data.relation_graph, hierarchy_name_list[-1], n.singular_name):
                    hierarchy_names.append(n)
            if len(hierarchy_names) == 0:
                raise KeyError(f"`{hierarchy_name_list[-1]}` does not have a factor `{singular_name}`")
        elif start is not None:
            if factor_name in start.factors:
                hierarchy_names = [start]
        if singular_name is None:
            singular_name = self.data.singular_name(factor_name)  # singular_name
        if hierarchy_names is None:
            hierarchy_names = self.data.factor_hierarchies[singular_name]
        if len(hierarchy_names) > 1:
            raise AmbiguousPathError(f"The factor {singular_name} is ambiguous. "
                                     f"{singular_name} has {len(hierarchy_names)} parents: {hierarchy_names}."
                                     f"Be explicit and choose one of them. "
                                     f"E.g. {hierarchy_names[0]}.{factor_name}")
        else:
            return hierarchy_names[0].singular_name.lower(), factor_name, singular_name

    def path(self, start, end) -> 'Path':
        raise NotImplementedError

    def _filter_by_boolean(self, parent, boolean):
        raise NotImplementedError

    def _equality(self, parent, other, negate=False):
        raise NotImplementedError

    def _compare(self, parent, other, operation):
        raise NotImplementedError

    def _combine(self, parent, other, operation):
        raise NotImplementedError



from collections import defaultdict
from functools import reduce, wraps
from operator import xor
from typing import Union, List, Dict, Optional

import networkx as nx
from networkx import OrderedDiGraph

from weaveio.writequery.base import BaseStatement, CypherVariable, CypherQuery, DerivedCypherVariable, CypherVariableItem, CypherData


def typeerror_is_false(func):
    @wraps(func)
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except TypeError:
            return False
    return inner


def sort_rooted_dag(graph):
    for n, d in graph.in_degree():
        if d == 0:
            break
    return nx.algorithms.traversal.dfs_tree(graph, n)


class Step:
    def __init__(self, direction: str, label: str = 'is_required_by', properties: Dict = None):
        if isinstance(direction, Step):
            self.direction = direction.direction
            self.label = direction.label
            self.properties = direction.properties
        elif direction in ['->', '<-', '-']:
            self.direction = direction
            self.label = label
            self.properties = properties
        else:
            raise ValueError(f"Direction {direction} is not supported")

    @typeerror_is_false
    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return self.direction == other.direction and self.properties == other.properties and self.label == other.label

    def __str__(self):
        if self.properties is None:
            mid = f'-[:{self.label}]-'
        else:
            mid = f'-[:{self.label} {self.properties}]-'
        if self.direction == '->':
            return f"{mid}>"
        elif self.direction == '<-':
            return f"<{mid}"
        else:
            return mid


class TraversalPath:
    def __init__(self, *path: Union[Step, str]):
        self._path = path
        self.nodes = []
        self.steps = []
        self.path = []
        self.end = CypherVariable(str(path[-1]))
        for i, entry in enumerate(path[:-1]):
            if not i % 2:  # even number
                step = Step(entry)
                self.steps.append(step)
                self.path.append(step)
            else:
                self.nodes.append(str(entry))
                self.path.append(f'(:{entry})')

    def __str__(self):
        end = f'({self.end}:{self.end.namehint})'
        return ''.join(map(str, self.path)) + end

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return self._path == other._path

    def __hash__(self):
        return hash(self._path)


class Action(BaseStatement):
    compare = None
    shape = None

    def to_cypher(self):
        raise NotImplementedError

    def __getitem__(self, item: CypherVariable):
        if isinstance(item, CypherVariableItem):
            return self.transformed_variables[item.parent].get(item.args)
        return self.transformed_variables[item]

    def __init__(self, input_variables: List[CypherVariable], output_variables: List[CypherVariable],
                 hidden_variables: List[CypherVariable] = None, transformed_variables: Dict[CypherVariable, CypherVariable] = None, target: CypherVariable = None):
        super(Action, self).__init__(input_variables, output_variables, hidden_variables)
        self.transformed_variables = {} if transformed_variables is None else transformed_variables
        self.target = target

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        base = set(self.input_variables) == set(other.input_variables) and self.__class__ is other.__class__
        for c in self.compare:
            selfthing = getattr(self, c, None)
            otherthing = getattr(other, c, None)
            base &= selfthing == otherthing
        return base

    def __hash__(self):
        base = reduce(xor, map(hash, [tuple(self.input_variables), self.__class__.__name__]))
        for c in self.compare:
            base ^= hash(getattr(self, c))
        return base

    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        return f'<{str(self)}>'


class EntryPoint(Action):
    compare = []

    def __init__(self):
        super().__init__([], [])

    def to_cypher(self):
        return ''

    def __str__(self):
        return 'EntryPoint'


class StartingPoint(Action):
    compare = ['label']

    def __init__(self, label):
        self.label = label
        self.hierarchy = CypherVariable(self.label)
        super().__init__([], [self.hierarchy], target=self.hierarchy)

    def to_cypher(self):
        return f"OPTIONAL MATCH ({self.hierarchy}:{self.label})"

    def __str__(self):
        return f'{self.label}'


class DataReference(Action):
    compare = ['hashes']

    def __init__(self, *data):
        import numpy as np
        self.hashes = reduce(xor, [hash(np.array(a).tobytes()) for a in data])
        ins = [CypherData(datum, delay=True) for datum in data]
        super().__init__(ins, [])

    def to_cypher(self):
        return '// added data here'

    def __str__(self):
        return 'DataReference'



# class OrderBy(Action):
#
#     def __init__(self, base_hierarchies, ):
#
#     def to_cypher(self):
#         collections = f'collect({self.sort_node}) as {self.collected_variable}'
#         with_statement = ', '.join(map(str, self.base_hierarchies + collections))
#         call_statement = f'with {self.collected_variable} UNWIND {self.collected_variable} as {self.sort_node} RETURN {self.sort_node}' \
#                          f' ORDER BY {self.sort_node}.{self.sort_property}'
#         return f"WITH {with_statement} + CALL {{ {call_statement} }}"


class Traversal(Action):
    """
    Traverse from one hierarchy level to another. This extends the branch and
    potentially increases the cardinality.
    Given a `source` branch and one or more `paths` of form (minnumber, maxnumber, direction),
     traverse to the nodes described by the `paths`.
    For example:
        >>> Traversal(branch, TraversalPath(['->', 'Exposure', '->', 'OB', '->', 'OBSpec']))
        results in `OPTIONAL MATCH (run)-[]->(:Exposure)-[]->(:OB)-[]->(obspec:OBSpec)`
    To traverse multiple paths at once, we use unions in a subquery
    """
    compare = ['paths', 'source']

    def __init__(self, source: CypherVariable, *paths: TraversalPath, name=None):
        if name is None:
            name = ''.join(p.end.namehint for p in paths)
        if len(paths) > 1:
            self.out = CypherVariable(name)
            outs = [p.end for p in paths] + [self.out]
        else:
            self.out = paths[0].end
            outs = [self.out]
        super(Traversal, self).__init__([source], outs, target=self.out)
        self.source = source
        self.paths = paths

    def to_cypher(self):
        lines = [f'OPTIONAL MATCH ({self.source}){p}' for p in self.paths]
        if len(self.paths) == 1:
            return lines[0]
        lines = '\n\nUNION\n\n'.join([f'\tWITH {self.source}\n\t{l}\n\tRETURN {path.end} as {self.out}' for l, path in zip(lines, self.paths)])
        return f"""CALL {{\n{lines}\n}}"""

    def __str__(self):
        return f'{self.source.namehint}->{self.out.namehint}'


class Return(Action):
    compare = ['branch', 'varnames']  # just compare input_variables

    def __init__(self, branch: 'Branch', *varnames):
        self.branch = branch
        self.varnames = varnames
        super(Return, self).__init__([branch.hierarchies[-1]], [])

    def to_cypher(self):
        proj = ', '.join([f'{self.branch.hierarchies[-1].get(v)}' for v in self.varnames])
        return f"RETURN {proj}"

    def __str__(self):
        return f'return {self.varnames}'


class Collection(Action):
    shape = 'rect'
    compare = ['_singles', '_multiples', '_reference']

    def __init__(self, reference: 'Branch', singles: List['Branch'], multiples: List['Branch']):
        self._singles = tuple(singles)
        self._multiples = tuple(multiples)
        self._reference = reference

        self.references = reference.hierarchies
        self.insingle_hierarchies = [x.hierarchies[-1] for x in singles]
        self.insingle_variables = [v for x in singles for v in x.variables]
        self.inmultiple_hierarchies = [x.hierarchies[-1] for x in multiples]
        self.inmultiple_variables = [v for x in multiples for v in x.variables]

        self.outsingle_hierarchies = [CypherVariable(s.namehint) for s in self.insingle_hierarchies]
        self.outsingle_variables = [CypherVariable(s.namehint) for s in self.insingle_variables]
        self.outmultiple_hierarchies = [CypherVariable(s.namehint+'_list') for s in self.inmultiple_hierarchies]
        self.outmultiple_variables = [CypherVariable(s.namehint+'_list') for s in self.inmultiple_variables]
        inputs = self.insingle_hierarchies + self.insingle_variables + self.inmultiple_variables + self.inmultiple_hierarchies
        outputs = self.outsingle_hierarchies + self.outsingle_variables + self.outmultiple_variables + self.outmultiple_hierarchies
        super().__init__(inputs + self.references, outputs, [], transformed_variables={i: o for i, o in zip(inputs, outputs)})

    def to_cypher(self):
        base = [f'{r}' for r in self.references + ['time0']]
        single_hierarchies = [f'coalesce({i}) as {o}' for i, o in zip(self.insingle_hierarchies, self.outsingle_hierarchies)]
        multiple_hierarchies = [f'collect({i}) as {o}' for i, o in zip(self.inmultiple_hierarchies, self.outmultiple_hierarchies)]
        single_variables = [f'coalesce({i}) as {o}' for i, o in zip(self.insingle_variables, self.outsingle_variables)]
        multiple_variables = [f'collect({i}) as {o}' for i, o in zip(self.inmultiple_variables, self.outmultiple_variables)]
        return 'WITH ' + ', '.join(base + single_hierarchies + single_variables + multiple_hierarchies + multiple_variables)

    def __str__(self):
        return f'collect'


class Operation(Action):
    compare = ['string_function', 'hashable_inputs']

    def __init__(self, string_function: str, **inputs):
        self.string_function = string_function
        self.output = CypherVariable('operation')
        self.inputs = inputs
        self.hashable_inputs = tuple(self.inputs.items())
        super().__init__(list(inputs.values()), [self.output], target=self.output)

    def to_cypher(self):
        return f"WITH *, {self.string_function.format(**self.inputs)} as {self.output_variables[0]}"

    def __str__(self):
        return self.string_function


class Filter(Operation):
    shape = 'diamond'
    def __init__(self, string_function, **inputs):
        super().__init__(string_function, **inputs)

    def to_cypher(self):
        return f"WHERE {self.string_function.format(**self.inputs)}"


# class Alignment(Action):
#     compare = ['branches', 'reference']
#     shape = 'house'
#
#     def __init__(self, reference: 'Branch', *branches: 'Branch'):
#         """
#         Collects and unwinds all variables/hierarchies that came after the reference branch
#         Persists all variables/hierarchies that came before the reference branch
#         """
#         self.reference = reference
#         self.branches = branches
#         base = tuple() if reference is None else (self.reference, )
#         ref_vars = [] if reference is None else reference.variables + reference.hierarchies
#
#         ins = []
#         before, after = set(), []
#         for branch in branches + base:
#             ins += branch.variables + branch.hierarchies
#             before |= {v for v in branch.variables + branch.hierarchies if v in ref_vars}
#             after += [v for v in branch.variables + branch.hierarchies if v not in ref_vars]
#         hidden = [CypherVariable(x.namehint+'_collected') for x in after]
#         outs = [CypherVariable(x.namehint+'_aligned') for x in after]
#         self.indexer = CypherVariable('i')
#         self.after = []
#         self.hidden = []
#         self.outs = []  # remove correlated duplicates
#         for a, h, o in zip(after, hidden, outs):
#             if a not in self.after:
#                 self.after.append(a)
#                 self.hidden.append(h)
#                 self.outs.append(o)
#         self.before = list(before)
#         self.output_variables = [o for a, o in zip(after, outs) if any(a in b.variables for b in branches)]
#         self.output_variables += reference.variables
#         self.output_hierarchies = [o for a, o in zip(after, outs) if any(a in b.hierarchies for b in branches)]
#         self.output_hierarchies += reference.hierarchies
#         transformed = {a: o for a, o in zip(after, outs)}
#         transformed.update({r: r for r in ref_vars})
#         super().__init__(ins, self.outs, self.hidden + [self.indexer], transformed)
#
#     def to_cypher(self):
#         base = ['time0'] + [str(b) for b in self.before]
#         base += [f'collect({i}) as {h}' for i, h in zip(self.after, self.hidden)]
#         unwind = f'UNWIND range(0, apoc.coll.max([x in {self.hidden} | size(x)])-1) as {self.indexer}'
#         get = [f'{h}[{self.indexer}] as {o}' for h, o in zip(self.hidden, self.outs)]
#         if len(self.after):
#             return f"WITH {', '.join(base)}\n{unwind}\nWITH *, {', '.join(get)}"
#         return f"WITH {', '.join(base)}"
#
#     def __str__(self):
#         return 'align'

class Alignment(Action):
    compare = ['reference', 'branches']

    def __init__(self, reference, *branches):
        self.reference = reference
        self.branches = branches
        super(Alignment, self).__init__([], [])

    def to_cypher(self):
        return ''

    def __str__(self):
        return 'align'


class Slice(Action):
    compare = []

    def __init__(self, slc):
        self.slc = slc
        self.skip = slc.start
        self.limit = slc.stop - slc.start
        super().__init__([], [], [])

    def __str__(self):
        return f'{self.slc}'

    def to_cypher(self):
        return f'WITH * SKIP {self.skip} LIMIT {self.limit}'


class Results(Action):
    compare = ['branches']

    def __init__(self, branch_attributes):
        self.branch_attributes = branch_attributes
        self.branches = tuple(branch_attributes.keys())
        ins = [j for i in self.branch_attributes.values() for j in i]
        super(Results, self).__init__(ins, [], [], {}, None)

    def to_cypher(self):
        return 'RETURN {}'.format(', '.join(map(str, self.input_variables)))

    def __str__(self):
        names = [i.namehint for i in self.input_variables]
        return 'return {}'.format(', '.join(names))



class BranchHandler:
    def __init__(self):
        self.graph = OrderedDiGraph()
        self.class_counter = defaultdict(int)
        self.entry = self.new(EntryPoint(), [], [], None, [], [], [])
        self.data_objects = {}

    def new(self, action: Action, accessible_parents: List['Branch'], inaccessible_parents: List['Branch'],
            current_hierarchy: Optional[CypherVariable], current_variables: List[CypherVariable],
            variables: List[CypherVariable], hierarchies: List[CypherVariable], name: str = None):
        parents = accessible_parents + inaccessible_parents
        parent_set = set(parents)
        successors = {s for p in parents for s in self.graph.successors(p) if set(self.graph.predecessors(s)) == parent_set}
        candidates = {s for s in successors if s.action == action}
        assert len(candidates) <= 1
        if candidates:
            return successors.pop()
        if name is None:
            self.class_counter[action.__class__] += 1
            name = action.__class__.__name__ + str(self.class_counter[action.__class__])
        instance = Branch(self, action, accessible_parents, inaccessible_parents, current_hierarchy, current_variables,
                          variables=variables, hierarchies=hierarchies, name=name)
        attrs = {}
        if action.shape is not None:
            attrs['shape'] = action.shape
        self.graph.add_node(instance, action=action, name=name, **attrs)
        for parent in parents:
            self.graph.add_edge(parent, instance, accessible=parent in accessible_parents)
        self.current_hierarchy = current_hierarchy
        return instance

    def begin(self, label):
        action = StartingPoint(label)
        return self.new(action, [self.entry], [], action.hierarchy, [action.hierarchy], [], [action.hierarchy])

    def relevant_graph(self, branch):
        return nx.subgraph_view(self.graph, lambda n: nx.has_path(self.graph, n, branch) or n is branch)

    def deepest_common_ancestor(self, *branches: 'Branch'):
        common = set()
        for i, branch in enumerate(branches):
            ancestors = set(nx.algorithms.dag.ancestors(self.graph, branch))
            if i == 0:
                common = ancestors
            else:
                common &= ancestors
        if not len(common):
            return None
        distances = [(sum(nx.shortest_path_length(self.graph, ancestor, b) for b in branches), ancestor) for ancestor in common]
        return distances[distances.index(min(distances, key=lambda x: x[0]))][1]


def plot(graph, fname):
    from networkx.drawing.nx_agraph import to_agraph
    A = to_agraph(graph)
    A.layout('dot')
    A.draw(fname)


class Branch:
    def __init__(self, handler: BranchHandler, action: Action,
                 accessible_parents: List['Branch'], inaccessible_parents: List['Branch'],
                 current_hierarchy: Optional[CypherVariable], current_variables: List[CypherVariable],
                 hierarchies: List[CypherVariable],
                 variables: List[CypherVariable], name: str = None):
        """
        A branch is an object that represents all the Actions (query statements) in a query.
        It contains both a node (Action) and references to all actions (other Branches) preceeding it (parents).
        If a branch is created in the same way more than once, only one object is actually instantiated, this is to optimize performance
        when writing to cypher query language. The user shouldn't care about this quirk.
        :param handler: The handler object which oversees query uniqueness for a particular dataset.
        :param action: The action that this branch will execute at the end. These are generally (but not always) Cypher statements.
        :param parents: The branches that must be executed before this one (i.e. the dependencies of this branch)
        """
        self.handler = handler
        self.action = action
        self.parents = accessible_parents + inaccessible_parents
        self.accessible_parents = accessible_parents
        self.inaccessible_parents = inaccessible_parents
        self.current_hierarchy = current_hierarchy
        self.current_variables = current_variables
        self.name = name
        self.hierarchies = hierarchies
        self.variables = variables
        self._relevant_graph = None
        self._accessible_graph = None
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = reduce(xor, map(hash, self.parents + [self.current_hierarchy, tuple(self.current_variables), self.action, self.handler]))
        return self._hash

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        return self.handler == other.handler and self.action == other.action and set(self.parents) == set(other.parents) \
               and self.current_hierarchy == other.current_hierarchy and set(self.current_variables) == set(other.current_variables)

    def __repr__(self):
        return f'<Branch {self.name}: {str(self.action)}>'

    @property
    def relevant_graph(self):
        if self._relevant_graph is None:
            self._relevant_graph = self.handler.relevant_graph(self)
        return self._relevant_graph

    @property
    def accessible_graph(self):
        if self._accessible_graph is None:
            g = nx.subgraph_view(self.relevant_graph, filter_edge=lambda a, b: self.relevant_graph.edges[(a, b)]['accessible'])
            self._accessible_graph = nx.subgraph_view(g, filter_node=lambda x: nx.has_path(g, x, self))
        return self._accessible_graph

    def iterdown(self, graph=None, start=None):
        if graph is None:
            graph = self.relevant_graph
        if start is None:
            start = self.handler.entry
        return nx.algorithms.traversal.dfs_tree(graph, start)

    def find_hierarchies(self):
        hierarchies = []
        for branch in self.iterdown(self.accessible_graph):
            if branch.current_hierarchy is not None:
                hierarchies.append(branch.current_hierarchy)
        return hierarchies

    def find_variables(self):
        variables = []
        for branch in self.iterdown(self.accessible_graph):
            variables += branch.current_variables
        return variables

    def find_hierarchy_branches(self):
        branches = []
        for branch in self.iterdown(self.accessible_graph):
            if branch.current_hierarchy is not None:
                branches.append(branch)
        return branches

    def add_data(self, *data) -> 'Branch':
        action = DataReference(*data)
        return self.handler.new(action, [self], [], current_hierarchy=None, current_variables=action.input_variables,
                         variables=self.variables+[action.input_variables], hierarchies=self.hierarchies)

    def traverse(self, *paths: TraversalPath) -> 'Branch':
        """
        Extend the branch from the most recent hierarchy to a new hierarchy(s) by walking along the paths
        If more than one path is given then we take the union of all the resultant nodes.
        """
        action = Traversal(self.find_hierarchies()[-1], *paths)
        return self.handler.new(action, [self], [], current_hierarchy=action.out, current_variables=[action.out],
                                variables=self.variables, hierarchies=self.hierarchies+[action.out])

    def align(self, branch: 'Branch') -> 'Branch':
        """
        Join branches into one, keeping the highest cardinality.
        This is used to to directly compare arrays:
            * ob1.runs == ob2.runs  (unequal sizes are zipped up together)
            * ob1.runs == run1  (array to single comparisons are left as is)
            * run1 == run2  (single to single comparisons are left as is)
        zip ups and unwinds take place relative to the branch's shared ancestor
        """
        action = Alignment(self, branch)
        return self.handler.new(action, [self, branch], [], None, current_variables=[],
                                variables=self.variables, hierarchies=self.hierarchies)

    def collect(self, singular: List['Branch'], multiple: List['Branch']) -> 'Branch':
        """
        Join branches into one, reducing the cardinality to this branch.
        `singular` contains branches that will be coalesced (i.e. only the first result is taken)
        `multiple` contains branches that will be collected (i.e. all results are presented in a list)
        This is used in predicate filters:
            ob1.runs[any(ob1.runs.l1singlespectra.snr > 10)]
            0. branch `ob1.runs` is created
            1. branch `ob1.runs.l1singlespectra.snr > 10` is created
            2. branch `ob1.runs.l1singlespectra.snr > 10` is collected with respect to `ob1.runs`
            3. A filter is applied on the collection at the `ob1.runs` level
        After a collection, only
        """
        action = Collection(self, singular, multiple)
        variables = action.outsingle_variables + action.outmultiple_variables
        hierarchies = action.outsingle_hierarchies + action.outmultiple_hierarchies
        return self.handler.new(action, [self], singular + multiple, None, variables + hierarchies,
                                variables=self.variables + variables, hierarchies=self.hierarchies + hierarchies)

    def operate(self, string_function, **inputs) -> 'Branch':
        """
        Adds a new variable to the namespace
        e.g. y = x*2 uses extant variable x to define a new variable y which is then subsequently accessible
        """
        # missing = [k for k, v in inputs.items() if getattr(v, 'parent', v) not in self.variables + self.hierarchies]
        # if missing:
        #     raise ValueError(f"inputs {missing} are not in scope for {self}")
        op = Operation(string_function, **inputs)
        return self.handler.new(op, [self], [], None, op.output_variables,
                                variables=self.variables + op.output_variables, hierarchies=self.hierarchies)

    def filter(self, logical_string, **boolean_variables: CypherVariable) -> 'Branch':
        """
        Reduces the cardinality of the branch by using a WHERE clause.
        .filter can only use available variables
        """
        # missing = [k for k, v in boolean_variables.items() if getattr(v, 'parent', v) not in self.variables + self.hierarchies]
        # if missing:
        #     raise ValueError(f"inputs {missing} are not in scope for {self}")
        action = Filter(logical_string, **boolean_variables)
        return self.handler.new(action, [self], [], None, [],
                                variables=self.variables, hierarchies=self.hierarchies)

    def results(self, branch_attributes: Dict['Branch', List[CypherVariable]]):
        """
        Returns the rows of results.
        All other branches which are not this branch will have their results collected
        """
        branch_attributes = {k: v if isinstance(v, (list, tuple)) else [v] for k, v in branch_attributes.items()}
        action = Results(branch_attributes)
        return self.handler.new(action, [self], [], None, [], self.variables, self.hierarchies)



from collections import defaultdict
from typing import Dict, Any, Union, List, Type

from weaveio.hierarchy import Hierarchy




def parse_apoc_tree(root_hierarchy: Type['Hierarchy'], root_id: Any, tree: Dict[str, Any],
                    data) -> Union['Hierarchy', List['Hierarchy']]:
    hierarchies = data.class_hierarchies
    subclassed = {k.__name__: len(k.__subclasses__()) for k in hierarchies.values()}
    inputs = defaultdict(list)
    for key, value in tree.items():
        if key.startswith('_') or key == 'id':
            continue
        elif isinstance(value, list):  # preceeding relationship
            value.sort(key=lambda x: x.pop(f'{key}.order', 0))
            for entry in value:
                names = [n for n in entry['_type'].split(':') if n in hierarchies]
                names.sort(key=lambda x: subclassed[x])
                name = names[0]
                # if issubclass(hierarchies[name], File):
                #     h = hierarchies[name](fname=entry['id'])
                h = parse_apoc_tree(hierarchies[name], entry['id'], entry, data)
                requirements = root_hierarchy.requirement_names()
                if h.singular_name in requirements:
                    inputs[h.singular_name] = h
                elif h.plural_name in requirements:
                    inputs[h.plural_name].append(h)
                else:
                    raise TypeError(f"{root_hierarchy} does not require {h}. This should not happen")
        elif isinstance(value, (int, float, str)):
            inputs[key.lower()] = value
        else:
            raise ValueError(f"Invalid json schema")
    h = root_hierarchy(**inputs)
    h.identifier = root_id
    h.add_parent_data(data)
    return h


from typing import List, Union

import numpy as np
import py2neo
from astropy.table import Table, Column, hstack

from weaveio.basequery.common import FrozenQuery, UnexpectedResult, NotYetImplementedError
from weaveio.basequery.tree import Branch
from weaveio.writequery import CypherVariable, CypherQuery


class FactorFrozenQuery(FrozenQuery):
    def __init__(self, handler, branch: Branch, factors: List[str], factor_variables: List[CypherVariable],
                 numbers: List[Union[int, None]], parent: FrozenQuery = None):
        super().__init__(handler, branch, parent)
        self.factors = factors
        self.factor_variables = factor_variables
        self.numbers = numbers

    def _prepare_query(self) -> CypherQuery:
        with super()._prepare_query() as query:
            return query.returns(*self.factor_variables)

    def __repr__(self):
        if isinstance(self.factors, tuple):
            factors = f'{self.factors}'
        else:
            factors = f'[{self.factors}]'
        return f'{self.parent}{factors}'

    def _post_process(self, result: py2neo.Cursor) -> Table:
        coldata = list(zip(*result.to_table()))
        columns = []
        for i, (name, number, data) in enumerate(zip(self.factors, self.numbers, map(list, coldata))):
            if number is None:
                #  make an unstructured list
                shape = len(data)
                dtype = object
                data[0].append('placeholder')  # to force astropy to pay attention!!!
                data.append(['placeholder'])
            elif number == 0:
                #  make a single scalar column
                shape = len(data)
                dtype = None
            else:
                # make a shaped vector Column
                shape = (len(data), number)
                dtype = None
            column = Column(data, name=name, shape=shape, length=len(data), dtype=dtype)
            if number is None:
                del column[0][-1]  # delete the placeholder
                column = column[:-1]
            columns.append(column)
        table = hstack(columns)
        return table


class SingleFactorFrozenQuery(FactorFrozenQuery):
    """A single factor of a single hierarchy instance"""

    def _post_process(self, result: py2neo.Cursor):
        t = super(SingleFactorFrozenQuery, self)._post_process(result)
        if len(t) != 1 or len(t.colnames) != 1:
            raise UnexpectedResult(f"Query raised a shape of {(len(t), len(t.colnames))} instead of (1, 1)")
        return t[t.colnames[0]][0]


class ColumnFactorFrozenQuery(FactorFrozenQuery):
    """A list of the same factor values for different hierarchy instances"""
    def _post_process(self, result: py2neo.Cursor):
        t = super(ColumnFactorFrozenQuery, self)._post_process(result)
        if len(t.colnames) != 1:
            raise UnexpectedResult(f"Query raised {len(t.colnames)} instead of (..., 1)")
        return t[t.colnames[0]].data


class TableFactorFrozenQuery(FactorFrozenQuery):
    """
    A matrix of different factors against different hierarchy instances
    This is only possible if the hierarchies each have only one of the factors
    """
    def __init__(self, handler, branch, factors, factor_variables, numbers, return_keys: List[str] = None, parent: 'FrozenQuery' = None):
        super().__init__(handler, branch, factors, factor_variables, numbers, parent)
        self.return_keys = return_keys

    def _prepare_query(self) -> CypherQuery:
        with super()._prepare_query() as query:
            variables = {k: v for k, v in zip(self.return_keys, self.factor_variables)}
            return query.returns(**variables)

    def _post_process(self, result):
        t = super(TableFactorFrozenQuery, self)._post_process(result)
        if self.return_keys is None:
            t = np.asarray([t[c].data for c in t.colnames]).T  # without column names
        return t

    def __getattr__(self, item):
        raise NotYetImplementedError


class RowFactorFrozenQuery(TableFactorFrozenQuery):
    """A list of different factors for one hierarchy"""
    def _post_process(self, result: py2neo.Cursor):
        df_or_array = super(RowFactorFrozenQuery, self)._post_process(result)
        assert len(df_or_array) == 1, f"Unexpected number of results ({len(df_or_array)} for {self}"
        return df_or_array[0]
